/*
 *  2017 by Ignacio Hernandez-Ros.
 *
 * This file is property of Reporting Standard S.L. and shall not be used outside the company unless you
 * have written permissions for different purposes.
 *
 * If you have received this file from Reporting Standard, you may have the following permissions:
 *
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2017 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 25 jun. 2017
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.IOException;
import java.net.URI;
import java.util.Iterator;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XMLDocumentImpl;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLSerializationException;
import com.ihr.xbrl.om.instance.XBRLContext;
import com.ihr.xbrl.om.instance.XBRLDimensionsContainer;
import com.ihr.xbrl.om.instance.XBRLEntity;
import com.ihr.xbrl.om.instance.XBRLFact;
import com.ihr.xbrl.om.instance.XBRLFactItem;
import com.ihr.xbrl.om.instance.XBRLFactNonNumeric;
import com.ihr.xbrl.om.instance.XBRLFactNumeric;
import com.ihr.xbrl.om.instance.XBRLFactTuple;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.instance.XBRLNumber;
import com.ihr.xbrl.om.instance.XBRLPeriod;
import com.ihr.xbrl.om.instance.XBRLScenario;
import com.ihr.xbrl.om.instance.XBRLUnit;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.ihr.xbrl.util.XmlFormater;

import net.sf.saxon.om.AtomicSequence;
import net.sf.saxon.om.Sequence;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmValue;
import net.sf.saxon.value.AtomicValue;
import net.sf.saxon.value.EmptySequence;
import net.sf.saxon.value.NumericValue;

/**
 * This class contains static methods to perform the comparison of two XBRL reports
 * using only the semantical information. For example a fact item with contextRef="u1" 
 * is semantically equivalent to another fact with contextRef="GBP" if both units exists
 * in its respective reports and are U-equal
 * 
 * @author Ignacio
 *
 */
public class CompareInstancesSemantically {
	
	private static final Logger log = Logger.getLogger(CompareInstancesSemantically.class);

	/**
	 * @param formulaProcessorGeneratedInstance
	 * @param instanceNode
	 * @param resultText 
	 * @return
	 */
	public static boolean compareInstances(XBRLInstance inputInstance, XdmNode instanceNode, StringBuffer resultText) {

		String xEqualMode = inputInstance.getContainer().getProperties().getProperty(DTSContainer.XEQUAL_MODE, DTSContainer.USE_XEQUAL_AS_XBRL_COMPLIANT);
		boolean changeValue = !xEqualMode.equals(DTSContainer.USE_XEQUAL_AS_FORMULA_COMPLIANT);
		if (changeValue)
			inputInstance.getContainer().getProperties().setProperty(DTSContainer.XEQUAL_MODE, DTSContainer.USE_XEQUAL_AS_FORMULA_COMPLIANT);

		// create a new instance in the DTS from fpInstance using nodes from instanceNode
		try {
			DTSContainer dts = inputInstance.getContainer();

			// obtain the instance in the conformace suite to compare to
			XBRLInstance cpInstance = null;
			try {
				cpInstance = new XBRLInstance(dts, instanceNode.asSource(), true);
			} catch (SaxonApiException e) {
				resultText.append("There is an error in the instance document in the conformance suite.\n"+instanceNode.getDocumentURI()+"\n"+e.getMessage());
				return false;
			}
			
			// Prepare the output instance to compare with
			URI inpInstanceURI = DTSContainer.getTempURL().resolve("tempInstance1.xbrl");
			inputInstance.setURI(inpInstanceURI);
			XMLDocumentImpl docInputInstance = inputInstance.generateDocNode(/*false, null*/);
			XBRLInstance fpInstance = null;
			try {
				XdmNode inputInstanceNode = docInputInstance.getXdmNode();
				fpInstance = new XBRLInstance(dts, inputInstanceNode.asSource(), true);
			} catch (SaxonApiException e) {
				resultText.append("There is an error in the generated instance document.\n"+instanceNode.getDocumentURI()+"\n"+e.getMessage());
				return false;
			}
			

			if (log.isTraceEnabled()) {
				XdmNode cpD = cpInstance.getDocument();
				XdmNode fpD = fpInstance.getDocument();
				log.trace("Comparing instances:");
				log.trace("Conformance suite instance:");
				log.trace(new String(XmlFormater.format(cpD), "UTF-8").toString());
				log.trace("Generated instance:");
				log.trace(new String(XmlFormater.format(fpD), "UTF-8").toString());
			}
			
			// check number of facts
			if (cpInstance.getInstanceRootNode().size() != fpInstance.getInstanceRootNode().size()) {
				resultText.append("Number of concepts in instances does not match.");
				return false;
			}
			
			// now compare the two instances...
			
			// fact comparison
			Iterator<XBRLFact> iterF = fpInstance.getInstanceRootNode().iterator();
			while (iterF.hasNext()) {
				XBRLFact fpFact = iterF.next();
				boolean bFound = false;
				Iterator<XBRLFact> iterF2 = cpInstance.getInstanceRootNode().iterator();
				while (iterF2.hasNext() && ! bFound) {
					XBRLFact cpFact = iterF2.next();
					
					if (compareTwoFacts(fpFact, cpFact, resultText)) {
						bFound = true;
						resultText.setLength(0);
					}
				}
				if (!bFound) {
					resultText.append("Fact "+fpFact.getNodeName().toString()+" not found.");
					return false;
				}
			}
			
			return true;
		} catch (IOException e) {
			e.printStackTrace();
		} catch (XBRLSerializationException e) {
			e.printStackTrace();
		} catch (DTSDiscoveryException e) {
			e.printStackTrace();
		} finally {
			if (changeValue)
				inputInstance.getContainer().getProperties().setProperty(DTSContainer.XEQUAL_MODE, xEqualMode);
		}
		return false;
	}
//	private static boolean compareInstances(XBRLInstance inputInstance, XdmNode instanceNode, StringBuffer resultText) {
//		// create a new instance in the DTS from fpInstance using nodes from instanceNode
//		try {
//			DTSContainer dts = inputInstance.getContainer();
//
//			// obtain the instance in the conformace suite to compare to
//			XBRLInstance cpInstance = null;
//			try {
//				cpInstance = new XBRLInstance(dts, instanceNode.asSource(), true);
//			} catch (SaxonApiException e) {
//				resultText.append("There is an error in the instance document in the conformance suite.\n"+instanceNode.getDocumentURI()+"\n"+e.getMessage());
//				return false;
//			}
//			
//			// Prepare the output instance to compare with
//			URI inpInstanceURI = DTSContainer.getTempURL().resolve("tempInstance1.xbrl");
//			inputInstance.setURI(inpInstanceURI);
//
//			XMLDocumentImpl docInputInstance = inputInstance.generateDocNode(false, null);
//			String ss = new String(XmlFormater.format(docInputInstance));
//			StreamSource ss2 = new StreamSource(new StringReader(ss));
//			ss2.setSystemId(instanceNode.asSource().getSystemId());
//			XBRLInstance fpInstance = new XBRLInstance(dts, ss2, true);
//			
////			cpInstance.updateXML(false);
////			fpInstance.updateXML(false);
//
//			if (log.isTraceEnabled()) {
//				XdmNode cpD = cpInstance.getDocument();
//				XdmNode fpD = fpInstance.getDocument();
//				log.trace("Comparing instances:");
//				log.trace("Conformance suite instance:");
//				log.trace(new String(XmlFormater.format(cpD), "UTF-8").toString());
//				log.trace("Generated instance:");
//				log.trace(new String(XmlFormater.format(fpD), "UTF-8").toString());
//			}
//			
//			// check number of facts
//			if (cpInstance.size() != fpInstance.size()) {
//				resultText.append("Number of concepts in instances does not match.");
//				return false;
//			}
//			
//			// now compare the two instances...
//			
//			// fact comparison
//			Iterator<XBRLFact> iterF = fpInstance.iterator();
//			while (iterF.hasNext()) {
//				XBRLFact fpFact = iterF.next();
//				boolean bFound = false;
//				Iterator<XBRLFact> iterF2 = cpInstance.iterator();
//				while (iterF2.hasNext() && ! bFound) {
//					XBRLFact cpFact = iterF2.next();
//					
//					if (compareTwoFacts(fpFact, cpFact, resultText)) {
//						bFound = true;
//						resultText.setLength(0);
//					}
//				}
//				if (!bFound) {
//					resultText.append("Fact "+fpFact.getNodeName().toString()+" not found.");
//					return false;
//				}
//			}
//			
//			return true;
//		} catch (SaxonApiException e) {
//			e.printStackTrace();
//		} catch (IOException e) {
//			e.printStackTrace();
//		} catch (XBRLSerializationException e) {
//			e.printStackTrace();
//		} catch (DTSDiscoveryException e) {
//			e.printStackTrace();
//		}
//		return false;
//	}

	/**
	 * Compare two facts and return true if they are semantically equivalent
	 * Note: semantically equivalent facts has not been defined in any XBRL specification.
	 * 
	 * @param fpFact
	 * @param cpFact
	 * @param resultText 
	 * @return boolean value
	 */
	private static boolean compareTwoFacts(XBRLFact fpFact, XBRLFact cpFact, StringBuffer resultText) {
		if (!fpFact.getNodeName().equals(cpFact.getNodeName()))
			return false;
		if (fpFact instanceof XBRLFactItem && cpFact instanceof XBRLFactItem)
			return compareTwoItemFacts((XBRLFactItem)fpFact, (XBRLFactItem)cpFact, resultText);
		else if (fpFact instanceof XBRLFactTuple && cpFact instanceof XBRLFactTuple)
			return compareTwoTupleFacts((XBRLFactTuple)fpFact, (XBRLFactTuple)cpFact, resultText);
		return true;
	}

	/**
	 * @param fpFact
	 * @param cpFact
	 * @param resultText 
	 * @return
	 */
	private static boolean compareTwoTupleFacts(XBRLFactTuple fpFact, XBRLFactTuple cpFact, StringBuffer resultText) {
		if (fpFact.size() != cpFact.size()) {
			resultText.append("tuple has different number of elements");
			return false;
		}
		Iterator<XBRLFact> iter1 = fpFact.iterator();
		Iterator<XBRLFact> iter2 = cpFact.iterator();
		while (iter1.hasNext()) {
			XBRLFact fact1 = iter1.next();
			XBRLFact fact2 = iter2.next();
			if (!compareTwoFacts(fact1,fact2,resultText))
				return false;
		}
		return false;
	}

	/**
	 * @param fpFact
	 * @param cpFact
	 * @param resultText 
	 * @return
	 */
	private static boolean compareTwoItemFacts(XBRLFactItem fpFact, XBRLFactItem cpFact, StringBuffer resultText) {
		// compare contexts		
		XBRLContext ctx1 = fpFact.getContext();
		XBRLContext ctx2 = cpFact.getContext();
		if (!compareTwoContexts(ctx1, ctx2, resultText)) {
			resultText.append(" not equal contexts");
			return false;
		}
		if (fpFact instanceof XBRLFactNumeric && cpFact instanceof XBRLFactNumeric)
			return compareTwoNumericFacts((XBRLFactNumeric)fpFact,(XBRLFactNumeric)cpFact,resultText);
		else if (fpFact instanceof XBRLFactNonNumeric && cpFact instanceof XBRLFactNonNumeric)
			return compareTwoNonNumericFacts((XBRLFactNonNumeric)fpFact,(XBRLFactNonNumeric)cpFact,resultText);
		return true;
	}

	/**
	 * @param ctx1
	 * @param ctx2
	 * @return
	 */
	private static boolean compareTwoContexts(XBRLContext ctx1, XBRLContext ctx2,StringBuffer resultText) {
//		return ctx1.equals(ctx2);
//		if (ctx1.hashCode() != ctx2.hashCode())
//			return false;
//		return true;
		
		// implement our own context comparison algorithm
		XBRLEntity entity1 = ctx1.getEntity();
		XBRLEntity entity2 = ctx2.getEntity();
		if (!compareTwoEntities(entity1, entity2, resultText)) {
			resultText.append(" not equal entities");
			return false;
		}
		
		XBRLPeriod p1 = ctx1.getPeriod();
		XBRLPeriod p2 = ctx2.getPeriod();
		if (!p1.equals(p2)) {
			resultText.append(" not equal periods");
			return false;
		}
		
		XBRLScenario s1 = ctx1.getScenario();
		XBRLScenario s2 = ctx2.getScenario();
		
		if (!compareTwoDimensionContainers(s1, s2, resultText))
			return false;
		return true;
	}

	/**
	 * @param entity1
	 * @param entity2
	 * @return
	 */
	private static boolean compareTwoEntities(XBRLEntity entity1, XBRLEntity entity2, StringBuffer sb) {
		if (!entity1.getScheme().equals(entity2.getScheme()))
			return false;
		if (!entity1.getIdentifier().equals(entity2.getIdentifier()))
			return false;	
		if (!compareTwoDimensionContainers(entity1.getSegment(), entity2.getSegment(), sb))
			return false;
		return true;
	}

	/**
	 * @param segment
	 * @param segment2
	 * @return
	 */
	private static boolean compareTwoDimensionContainers(
			XBRLDimensionsContainer dc1,
			XBRLDimensionsContainer dc2,
			StringBuffer sb) {
		if ((dc1 == null || dc1.isEmpty()) && (dc2 == null || dc2.isEmpty()))
			return true;
		if (dc1.getNumberOfElements() != dc2.getNumberOfElements()) {
			sb.append("number of children elements in the context container does not match");
			return false;
		}
		Iterator<XMLFragment> iter1 = dc1.getElements();
		int position = 0;
		int foundAtPosition = -1;
		while (iter1.hasNext()) {
			XMLFragment dimElement = iter1.next();
			// three possibilities: (1) a typedDimension container, (2) an explicitDimension container, (3) a node
			if (XDTProcessor.typedMemberContainerQName.equals(dimElement.getNodeName())) {
				// search for the same node in the target container. position is not important
				if (findElement(dimElement,dc2,0) < 0)
					return false;
			} else if (XDTProcessor.explicitMemberContainerQName.equals(dimElement.getNodeName())) {
				// search for the same node in the target container. position is not important
				if (findElement(dimElement,dc2,0) < 0)
					return false;
			} else {				
				// a node
				foundAtPosition = findElement(dimElement,dc2,position);
				// search for the same node in the target container. position is important.
				if (foundAtPosition == -1)
					return false;
				position = foundAtPosition+1; // first node to compare on next loop
			}
		}
		return true;
	}

	/**
	 * @param dimElement
	 * @param dc2
	 * @param position
	 * @return
	 */
	private static int findElement(XMLFragment el,	XBRLDimensionsContainer dc2, int position) {
		for (int i=position; i < dc2.getNumberOfElements(); ++i) {
			XMLFragment child = dc2.getNthElement(i);
//			if (child.equals(el))
//				return i;
			if (child.s_equal2(el))
				return i;
		}
		return -1;
	}

	/**
	 * @param fpFact
	 * @param cpFact
	 * @param resultText 
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private static boolean compareTwoNonNumericFacts(XBRLFactNonNumeric fpFact, XBRLFactNonNumeric cpFact, StringBuffer resultText) {
		XdmValue v1 = fpFact.getValue();
		XdmValue v2 = cpFact.getValue();
		Sequence vr1 = v1.getUnderlyingValue();
		Sequence vr2 = v2.getUnderlyingValue();

		if (vr1 instanceof AtomicValue && vr2 instanceof AtomicValue)
			return ((AtomicValue)vr1).getSchemaComparable().compareTo(((AtomicValue)vr2).getSchemaComparable()) == 0;
		if (vr1 == EmptySequence.getInstance() && vr2 == EmptySequence.getInstance())
			return true;
		resultText.append("not equal fact values");
		return false;
	}

	/**
	 * @param unit
	 * @param unit2
	 * @return
	 */
	private static boolean compareTwoUnits(XBRLUnit unit1, XBRLUnit unit2) {
		if (unit1.hashCode() != unit1.hashCode())
			return false;
		return true;
	}

	/**
	 * @param fpFact
	 * @param cpFact
	 * @param resultText 
	 * @return
	 */
	private static boolean compareTwoNumericFacts(XBRLFactNumeric fpFact, XBRLFactNumeric cpFact, StringBuffer resultText) {
		if (!compareTwoUnits(fpFact.getUnit(),cpFact.getUnit())) {
			resultText.append("not equal units");
			return false;
		}
		if (!compareAccuracy(fpFact,cpFact)) {
			resultText.append("not equal accuracy");
			return false;
		}
		if (cpFact.isNil())
			return fpFact.isNil();
		else if (fpFact.isNil())
			return false;
		XBRLNumber v1 = cpFact.getNumericValue();
		XBRLNumber v2 = fpFact.getNumericValue();
		if (v1 == null)
			return v2 == null;
		if (v2 == null)
			return v1 == null;
//		if (v1 instanceof XBRLNumericValue && v2 instanceof XBRLNumericValue) {
//			XBRLNumericValue nv1 = (XBRLNumericValue)v1;
//			XBRLNumericValue nv2 = (XBRLNumericValue)v2;
//			
//			return nv1.getBigDecimal().equals(nv2.getBigDecimal());
//		}
		if (cpFact.isNumber() && fpFact.isNumber()) {
			return cpFact.getBigDecimal().equals(fpFact.getBigDecimal());
		}
		
		Sequence val1 = cpFact.getValue().getUnderlyingValue();
		Sequence val2 = fpFact.getValue().getUnderlyingValue();
		if (val1 == EmptySequence.getInstance() && val2 == EmptySequence.getInstance())
			return true;
		if (val1 instanceof NumericValue && (!(val2 instanceof NumericValue) || !val1.equals(val2))) {
			resultText.append("not equal fact values");
			return false;
		}
		if (val2 instanceof NumericValue && (!(val1 instanceof NumericValue) || !val2.equals(val1))) {
			resultText.append("not equal fact values");
			return false;
		}
		if (!((AtomicSequence) val2).getStringValue().equals(((AtomicSequence) val1).getStringValue())) {
			resultText.append("not equal fact values");
			return false;
		}
		return true;
	}

	/**
	 * @param fpFact
	 * @param cpFact
	 * @return
	 */
	private static boolean compareAccuracy(XBRLFactNumeric fpFact,	XBRLFactNumeric cpFact) {
		String decimals1 = fpFact.getAttributeValue(new QName("decimals"));
		String precision1 = fpFact.getAttributeValue(new QName("precision"));
		String decimals2 = cpFact.getAttributeValue(new QName("decimals"));
		String precision2 = cpFact.getAttributeValue(new QName("precision"));
		if (decimals1 != null && !decimals1.equals(decimals2))
			return false;
		if (precision1 != null && !precision1.equals(precision2))
			return false;
		return true;
	}

}
