/*
 *  2007 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2007 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 12/05/2007
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Properties;

import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.s9api.DocumentBuilder;
import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XPathCompiler;
import net.sf.saxon.s9api.XPathSelector;
import net.sf.saxon.s9api.XdmItem;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmSequenceIterator;
import net.sf.saxon.s9api.XdmValue;
import net.sf.saxon.value.QNameValue;

import org.jdom.Attribute;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.XBRLDocument;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLInitializationException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.ihr.xbrl.om.xdt.errors.XDTValidationException;
import com.ihr.xbrl.util.XMLTools;
import com.rs.lic.LicenseVerificationException;

/**
 * @author Ignacio
 *
 */
public class LaunchDimensionsConformanceSuite {

	/**
	 * This is a map between errors declared in the conformance suite and errors generated by the dimensions
	 * processor. Note the test case error condition does not change, if a test case must raise an error, an
	 * error is raised. The difference is the error code detected. The existence of this table means some
	 * test cases contains two errors rather than just one and our processor detects the other error first
	 */
	private static HashMap<Integer,QName> replaceErrors = new HashMap<>();
	
	static {
		replaceErrors.put(75, new QName(XDTValidationException.XBRLDTE,"DimensionDomainTargetError"));
		replaceErrors.put(76, new QName(XDTValidationException.XBRLDTE,"DimensionDomainTargetError"));
		replaceErrors.put(77, new QName(XDTValidationException.XBRLDTE,"DimensionDomainTargetError"));
		replaceErrors.put(78, new QName(XDTValidationException.XBRLDTE,"DimensionDomainTargetError"));
		replaceErrors.put(81, new QName(XDTValidationException.XBRLDTE,"DimensionDomainTargetError"));
	}
	
	private static final String XHTMLNS = "http://www.w3.org/1999/xhtml";
	
	public static void usage() {
		System.out.println(
			"Usage: LaunchDimensionsConformanceSuite {-f {Conformance suite file}} [-m] [-x {exclusion pattern}] [-o {Output file}] [-t {test number,[test number]}]");
		System.exit(-1);
	}
	
	/**
	 * reads arguments starting at argument i. if it is a number then add the number to the list of
	 * test cases to process. continue until a non comma is found.
	 * 
	 * @param i
	 * @returns the next parameter to read
	 */
	private static int readList(int i, boolean bSkipComma, String args[], HashSet<Integer> hs) {
		if (args.length == i) {
			return i;
		}
		String whatToProcess = args[i].trim();
		if (!bSkipComma) {
			if (whatToProcess.startsWith(","))
				whatToProcess = whatToProcess.substring(1).trim();
			
			if (whatToProcess.isEmpty()) {
				return readList(++i,true,args,hs);
			}
		}
		try {
			StringTokenIterator sti = new StringTokenIterator(whatToProcess, ",");
			Item itm = null;
			while ((itm = sti.next())!= null) {
				int test = Integer.parseInt(itm.getStringValue());
				hs.add(test);
			}
		} catch (NumberFormatException e) {
			return i;
		}
		return readList(++i,true,args,hs);
	}

	/**
	 * @param args
	 * 
	 * args[0] = xbrl.xml
	 * args[1] = pattern to exclude tests
	 */
	public static void main(String[] args) {
		try {
			String confSuiteFile = null;
			String excludePatterns = null;
			String outputFile = null;
			boolean bShowMemoryUsage = false;
			
			// XBRL Processor properties
			Properties props = new Properties();
			String processors = XBRLCoreProcessor.PROCESSORNAME+","+XDTProcessor.PROCESSORNAME;
			props.setProperty(DTSContainer.PROCESSORS_SEQUENCE, processors);
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			HashSet<Integer> hsIncludedTests = null;
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-x".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					excludePatterns = args[i];
				} else if ("-t".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					if (hsIncludedTests == null)
						hsIncludedTests = new HashSet<Integer>();
					i = readList(i,false,args,hsIncludedTests);
					continue;
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				}
			}
			
			if (confSuiteFile == null)
				usage();
			
			// Load the conformance suite
			Processor proc = XMLTools.newProcessor(false);
			
			XPathCompiler xp = proc.newXPathCompiler();
			xp.declareNamespace("conf", "http://xbrl.org/2005/conformance");
			
			XPathSelector selTests = xp.compile("for $i in testcases/testcase return resolve-uri($i/@uri,base-uri($i))").load();
			XPathSelector selDesc = xp.compile("/conf:testcase/@description").load();
			XPathSelector selVar = xp.compile("/conf:testcase/conf:variation").load();
			XPathSelector selId = xp.compile("@id").load();
			XPathSelector selDesc2 = xp.compile("conf:description/text()").load();
			XPathSelector selFile = xp.compile("for $i in conf:data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))").load();
			XPathSelector selValues = xp.compile("count(conf:result/(conf:error|conf:warning|conf:inconsistency))").load();
			XPathSelector selRes = xp.compile("for $i in conf:result/conf:error return resolve-QName($i,$i)").load();
			
			StreamSource fs = new StreamSource(new File(confSuiteFile));
			DocumentBuilder builder = XMLTools.newDocumentBuilder(proc, TreeModel.TINY_TREE, null);
			XdmNode xbrl = builder.build(fs);

			// The output will be sent to a JDom document
			Document page = new Document();
			page.setDocType(new DocType("html","-//W3C//DTD XHTML 1.0 Strict//EN","http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"));
			Element html = new Element("html",XHTMLNS);
			page.addContent(html);
			Element head = new Element("head",XHTMLNS);
			html.addContent(head);
			Element title = new Element("title",XHTMLNS);
			title.setText("Compliance report with the Dimensions Conformance suite");
			head.addContent(title);
			Element style = new Element("style",XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type","text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body",XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2",XHTMLNS);
			pageH2.setText("Compliance report with the Dimensions Conformance suite");
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3",XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table",XHTMLNS);
			Attribute border1 = new Attribute("border","1");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			selTests.setContextItem(xbrl);
			XdmValue lst = selTests.evaluate();
			XdmSequenceIterator iter = lst.iterator();
			long now = System.currentTimeMillis();
			while (iter.hasNext()) {
				String strURI = iter.next().getStringValue();
				URI uri = new URI(strURI);
				StreamSource fstc = new StreamSource(uri.toURL().openStream());
				fstc.setSystemId(strURI);
				DocumentBuilder builder2 = XMLTools.newDocumentBuilder(proc, TreeModel.TINY_TREE, null);
				XdmNode testcases = builder2.build(fstc);
				selDesc.setContextItem(testcases);
				String description = selDesc.evaluateSingle().getStringValue();
				Element tr01 = new Element("tr",XHTMLNS);
				outTable.addContent(tr01);
				Element td01 = new Element("td",XHTMLNS);
				tr01.addContent(td01);
				td01.setText(description);
				selVar.setContextItem(testcases);
				XdmValue vars = selVar.evaluate();
				XdmSequenceIterator iterVars = vars.iterator();
				while (iterVars.hasNext()) {
					Element tr02 = new Element("tr",XHTMLNS);
					outTable.addContent(tr02);
					Element td02 = new Element("td",XHTMLNS);
					tr02.addContent(td02);
					Element table02 = new Element("table",XHTMLNS);
					td02.addContent(table02);
					Attribute at02 = new Attribute("border","1");
					table02.setAttribute(at02);
					Attribute at03 = new Attribute("width","100%");
					table02.setAttribute(at03);
					Element tr03 = new Element("tr",XHTMLNS);
					table02.addContent(tr03);
					XdmNode var = (XdmNode)iterVars.next();
					selId.setContextItem(var);
					XdmItem idItem = selId.evaluateSingle();
					String id = "";
					if (idItem != null)
						idItem.getStringValue();
					selDesc2.setContextItem(var);
					XdmItem descItem = selDesc2.evaluateSingle();
					String desc = "";
					if (descItem != null)
						desc = descItem.getStringValue();
					Element td03 = new Element("td",XHTMLNS);
					tr03.addContent(td03);
					td03.addContent(new Text("Id: "+id+" n:"+(testNumber+countSkiped+1)));
					Element td04 = new Element("td",XHTMLNS);
					td04.setText("Description: "+desc);
					tr03.addContent(td04);
					Element tr04 = new Element("tr",XHTMLNS);
					table02.addContent(tr04);
					selFile.setContextItem(var);
					String varF = selFile.evaluateSingle().getStringValue();
					if ((excludePatterns != null && varF.matches(excludePatterns)) || (hsIncludedTests != null && !hsIncludedTests.contains(testNumber+countSkiped+1))) {
						Element td05 = new Element("td",XHTMLNS);
						td05.setText("Execution excluded");
						tr04.addContent(td05);
						countSkiped++;
					} else {
						testNumber++;

						TestCaseDt tc = new TestCaseDt(bShowMemoryUsage,testNumber,countSkiped, props,varF);
						Thread th = new Thread(tc);
						th.start();
						try {
							th.join();
						} catch (InterruptedException e) {
						}
							
//							XBRLCoreProcessor xproc = XBRLCoreProcessor.create(dts);							
//							report = xproc.validate();
//							report.throwExceptions();
//							XDTProcessor xdt = XDTProcessor.create(dts);
//							report = xdt.validate();

						if (bShowMemoryUsage) {
							System.gc();
							long total = Runtime.getRuntime().totalMemory();
							long free = Runtime.getRuntime().freeMemory();
							System.out.println("Test number: "+testNumber+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
						}
						selValues.setContextItem(var);
						Double values = new Double(selValues.evaluateSingle().getStringValue());
						QName errors[] = new QName[values.intValue()];
						if (values.doubleValue() > 0) {
							selRes.setContextItem(var);
							XdmValue lstResNode = selRes.evaluate();
							XdmSequenceIterator iterLst = lstResNode.iterator();
							int i=0;
							while (iterLst.hasNext()) {
								QNameValue q = (QNameValue)iterLst.next().getUnderlyingValue();
								QName replaceWith = replaceErrors.get(testNumber+countSkiped);
								if (replaceWith != null)
									q = XMLTools.qnameValue(replaceWith);
								errors[i] = new QName(q.getNamespaceURI(),q.getLocalName());
								++i;
							}
						}
						if (errors.length > 0 && tc.bError) {
							boolean bShowError = true;
							for (int i=0; i < errors.length ; ++i) {
								if (errors[i].equals(tc.qError)) {
									Element td05 = new Element("td",XHTMLNS);
									Attribute at06 = new Attribute("class","testOK");
									td05.setAttribute(at06);
									td05.setText("OK. Expected invalid - result is invalid - right error detected");
									tr04.addContent(td05);
									countOk++;
									bShowError = false;
									break;
								}
							}
							if (bShowError) {
								Element td05 = new Element("td",XHTMLNS);
								Attribute at06 = new Attribute("class","testFail");
								td05.setAttribute(at06);
								td05.setText("NOT PASS. Expected invalid - result is invalid - wrong error detected.");
								tr04.addContent(td05);
								countError++;
							}
						} else if (errors.length > 0 && !tc.bError) {
							Element td05 = new Element("td",XHTMLNS);
							Attribute at06 = new Attribute("class","testFail");
							td05.setAttribute(at06);
							td05.setText("NOT PASS. Expected invalid - result is valid");
							tr04.addContent(td05);
							countError++;
						} else if (errors.length == 0 && tc.bError) {
							Element td05 = new Element("td",XHTMLNS);
							Attribute at06 = new Attribute("class","testFail");
							td05.setAttribute(at06);
							td05.setText("NOT PASS. Expected valid - result is invalid");
							tr04.addContent(td05);
							countError++;
						} else if (errors.length == 0 && !tc.bError) {
							Element td05 = new Element("td",XHTMLNS);
							Attribute at06 = new Attribute("class","testOK");
							td05.setAttribute(at06);
							td05.setText("OK. Expected valid - result is valid");
							tr04.addContent(td05);
							countOk++;
						} else {
							Element td05 = new Element("td",XHTMLNS);
							td05.setText("UNKNOWN. Expected result is "+tc.bError);
							tr04.addContent(td05);
							countError++;
						}
						Element td06 = new Element("td",XHTMLNS);
						td06.setText(tc.result.toString());
						tr04.addContent(td06);
						Element tr05 = new Element("tr",XHTMLNS);
						table02.addContent(tr05);
						Element td07 = new Element("td",XHTMLNS);
						Attribute at07 = new Attribute("colspan","2");
						td07.setAttribute(at07);
						td07.setText(varF);
						tr05.addContent(td07);
					}
				}
			}
			Element table03 = new Element("table",XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr",XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td",XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td",XHTMLNS);
			tr05.addContent(td08);
			td08.setText((new Integer(countError)).toString());
			Element tr06 = new Element("tr",XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td",XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td",XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr",XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td",XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td",XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());

			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else {
				System.out.println(s);
			}
			long now2 = System.currentTimeMillis();
			System.out.println("Total execution time "+(now2-now)/1000+"s");
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}
}

class TestCaseDt implements Runnable {
	
	StringBuffer result = new StringBuffer();
	boolean bError = false;
	private Properties props = null;
	String varF;
	URI uriF;
	boolean bShowMemoryUsage;
	int testNumber;
	int countSkiped;
	QName qError;
	
	TestCaseDt(boolean bShowMemoryUsage, int testNumber, int countSkiped, Properties props,String varF) throws URISyntaxException {
		this.bShowMemoryUsage = bShowMemoryUsage;
		this.testNumber = testNumber;
		this.countSkiped = countSkiped;
		this.props = props;
		this.varF = varF;
		uriF = new URI(varF);
	}

	@Override
	public void run() {
		DTSContainer dts = null;
		XBRLDocument doc = null;
		XBRLValidationReport report = null;
		
		try {
			if (bShowMemoryUsage) {
				long total = Runtime.getRuntime().totalMemory();
				long free = Runtime.getRuntime().freeMemory();
				System.out.println("Test number: "+testNumber+" Memory at 1: T="+total+" F="+free+" U="+(total-free)+" L=\""+varF+"\"");
			}
			dts = DTSContainer.newEmptyContainer(props);
			doc = dts.load(uriF);
			
			report = dts.validate();
			report.throwExceptions();
			//doc = null;
		} catch (XDTValidationException e) {
			bError = true;
			result.append(" ");
			result.append(e.getClass().getCanonicalName());
			result.append(" ");
			result.append(e.getLocalizedMessage());
			result.append(e.getQNameError().toString());
			result.append("\n");
			qError = e.getQNameError();
		} catch (XBRLValidationException e) {
			bError = true;
			result.append(" ");
			result.append(e.getClass().getCanonicalName());
			result.append(" ");
			result.append(e.getLocalizedMessage());
			if (e.getQNameError() != null) {
				result.append(e.getQNameError().toString());
				result.append("\n");
				qError = e.getQNameError();
			}
		} catch (DTSDiscoveryException e) {
			bError = true;
			result.append(" ");
			result.append(e.getClass().getCanonicalName());
			result.append(" ");
			result.append(e.getLocalizedMessage());
			result.append("\n");
			qError = new QName("http://www.xbrl.org/2005/genericXmlSchemaError","XmlSchemaError");
		} catch (LicenseVerificationException e) {
			e.printStackTrace();
			System.exit(-1);
		} catch (XBRLInitializationException e) {
			e.printStackTrace();
			System.exit(-1);
		}
		try {
			if (doc != null)
				doc.clear();
			if (dts != null)
				dts.clear();
			if (report != null)
				report.clear();
		} catch (Throwable e) {
			System.out.println("Exception thrown during memory cleanup on test #"+(testNumber+countSkiped)+"!");
		}
		doc = null;
		dts = null;
		report = null;
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
