package com.ihr.xbrl.sample;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Properties;
import java.util.Vector;

import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.jdom.Attribute;
import org.jdom.Content;
import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.Text;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.HasCreateFormulaParameterResolver;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLAssertionResult;
import com.ihr.xbrl.om.errors.XBRLInitializationException;
import com.ihr.xbrl.om.errors.XBRLRuntimeException;
import com.ihr.xbrl.om.errors.XBRLSerializationException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.formula.XBRLFormulaProcessor;
import com.ihr.xbrl.om.formula.errors.XFValidationException;
import com.ihr.xbrl.om.formula.functions.errors.XFException;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.ihr.xbrl.om.xdt.errors.XDTValidationException;
import com.ihr.xbrl.util.GeneralIterable;
import com.ihr.xbrl.util.XMLTools;
import com.ihr.xbrl.util.XmlFormater;
import com.rs.lic.LicenseVerificationException;

import net.sf.saxon.expr.StringTokenIterator;
import net.sf.saxon.om.AtomicSequence;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.s9api.DocumentBuilder;
import net.sf.saxon.s9api.Processor;
import net.sf.saxon.s9api.SaxonApiException;
import net.sf.saxon.s9api.XPathCompiler;
import net.sf.saxon.s9api.XPathExecutable;
import net.sf.saxon.s9api.XPathSelector;
import net.sf.saxon.s9api.XdmItem;
import net.sf.saxon.s9api.XdmNode;
import net.sf.saxon.s9api.XdmValue;
import net.sf.saxon.trans.XPathException;

public class LaunchFormulaConformanceSuiteXHTML {
	
	/**
	 * The formula conformance suite contains wrong errors when evaluating input parameters. This table
	 * replaces the wrong errors with the XPath original error that shall be raised by an XPath processor
	 * and not a custom error defined ad-hoc for things that are XPath errors
	 */
	private static HashMap<Integer,QName> replaceErrors = new HashMap<>();
	
	/**
	 * Some test cases in the conformance suite contains calculation inconsistencies but what is checked
	 * is number of assertions satisfied or not satisfied. For this cases raising exceptions when
	 * inconsistencies are found shall be disabled
	 */
	private static HashSet<Integer> allowInconsistencies = new HashSet<>();
	
	/**
	 * Some test cases might require special properties to the creation of the DTSContainer instance
	 */
	private static HashMap<Integer,Properties> testcaseProcessorProperties = new HashMap<>();
	
	static {
		allowInconsistencies.add(642);
		allowInconsistencies.add(643);
		allowInconsistencies.add(645);
		
//		replaceErrors.put(576, new QName("http://www.w3.org/2005/xqt-errors","FORG0006"));
//		replaceErrors.put(618, new QName("http://www.w3.org/2005/xqt-errors","XPTY0004"));
//		replaceErrors.put(619, new QName("http://www.w3.org/2005/xqt-errors","XPTY0004"));
//		replaceErrors.put(620, new QName("http://www.w3.org/2005/xqt-errors","XPTY0004"));
	}
	
	/**
	 * @author Ignacio
	 *
	 */
	public static class ConformanceParametersFactory implements HasCreateFormulaParameterResolver {
		
		private XdmValue value;

		/**
		 * @param evaluate
		 */
		public ConformanceParametersFactory(XdmValue value) {
			this.value = value;
		}

		/* (non-Javadoc)
		 * @see com.ihr.xbrl.om.HasCreateFormulaParameterResolver#createAndRegisterFormulaParameterResolver(com.ihr.xbrl.om.DTSContainer)
		 */
		@Override
		public void createAndRegisterFormulaParameterResolver(DTSContainer dts) {
			ConformanceParameters ccf = new ConformanceParameters(value);			
			dts.installFormulaParameterResolver(ccf);
			value = null;
		}
	}

	private static final String XHTMLNS = "http://www.w3.org/1999/xhtml";	
	private static final String CONFORMANCE_NAMESPACE = "http://xbrl.org/2008/conformance";
	
	private static final net.sf.saxon.s9api.QName errorQName = new net.sf.saxon.s9api.QName(CONFORMANCE_NAMESPACE,"c:error");
	private static final net.sf.saxon.s9api.QName assertionTestQName = new net.sf.saxon.s9api.QName(CONFORMANCE_NAMESPACE,"c:assertionTests");
	private static final net.sf.saxon.s9api.QName instanceTestQName = new net.sf.saxon.s9api.QName(CONFORMANCE_NAMESPACE,"c:instance");
	
	private static final Logger log = Logger.getLogger(LaunchConformanceSuiteXHTML.class);
	
	public static void usage() {
		System.out.println("Usage: LaunchFormulaConformanceSuiteXHTML {-f {Conformance suite file}} [-m] [-x {exclusion pattern}] [-r {starting test case number} {end test case number}] [-n {test case number}] [-o {Output file}] [-d DEBUG_LEVEL]\n" +
				"DEBUG LEVEL can be OFF,TRACE,... and applies to com.ihr.xbrl.om.formula");
		System.exit(-1);
	}
	
	@SuppressWarnings("unused")
	private static boolean debug = false;
		 		
	/**
	 * one error code should be generated
	 */
	private static int ERROR_CODE = 0x01;
		
	/**
	 * formulas that returns an instance document to compare
	 */
	private static int FORMULA = 0x02;
		
	/**
	 * Assertion test that contains number of satisfied and not satisfied tests
	 */
	private static int ASSERTION = 0x04;
	
	/**
	 * This code is used when the test case output contains both, the assertion test and one instance document
	 */
	private static int ASSERTION_PLUS_INSTANCE = FORMULA | ASSERTION;
	
	/**
	 * Valid (mostly for generic linkbases)
	 */
	private static int VALID = 0x08;
		
	/**
	 * Invalid (mostly for generic linkbases)
	 */
	private static int INVALID = 0x10;
	
	/**
	 * not specified or unknown
	 */
	private static int UNKNOWN = 0x000;

	/**
	 * @param i
	 * @param args
	 * @param dtsProps 
	 * @return
	 */
	private static int readProperties(int i, String[] args, Properties dtsProps) {
		if (args.length == i)
			return i;
		StringBuffer sb = new StringBuffer();
		int j=i;
		for (;j<args.length;++j) {
			String argument = args[j].trim();
			if (argument.startsWith("-")) {
				break;
			} else {
				sb.append(argument);
			}
		}		
		i = j-1;
		if (sb.length() == 0)
			usage();
		StringTokenIterator sti = new StringTokenIterator(sb.toString(), ",");
		Item itm = null;
		while ((itm = sti.next()) != null) {
			// find equal sign
			String token = itm.getStringValue();
			int idx = token.indexOf("=");
			if (idx < 0)
				usage();
			String propName = token.substring(0, idx);
			String propValue = token.substring(idx+1);
			dtsProps.setProperty(propName, propValue);
		}
		return i;
	}
	
	/**
	 * @param args
	 * 
	 * args[0] = xbrl.xml
	 * args[1] = pattern to exclude tests
	 */
	public static void main(String[] args) {
		try {
			LinkedList<Integer> errors = new LinkedList<>();
			String confSuiteFile = null;
			String excludePatterns = null;
			String outputFile = null;
			String debugLevel = "OFF";
			boolean bShowMemoryUsage = false;
			
			int countError = 0;
			int countOk = 0;
			int countSkiped = 0;
			int testNumber = 0;
			int startAt = 0;
			int endsAt = Integer.MAX_VALUE;
			int testType = UNKNOWN;
			
			ArrayList<Integer> testsToRun = new ArrayList<Integer>();
			Properties dtsProps = new Properties();
			
			// parse arguments
			for (int i=0; i < args.length; ++i) {
				if ("-f".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					confSuiteFile = args[i];
				} else if ("-m".equals(args[i])) {
					bShowMemoryUsage = true;
				} else if ("-x".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					excludePatterns = args[i];
				} else if ("-o".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					outputFile = args[i];
				} else if ("-r".equals(args[i])) {
					if (args.length < i+1)
						usage();
					++i;
					startAt = Integer.parseInt(args[i])-1;
					++i;
					endsAt = Integer.parseInt(args[i])-1;
				} else if ("-p".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					dtsProps.clear();
					i = readProperties(i,args,dtsProps);
					continue;
				} else if ("-d".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					debugLevel = args[i];					
				} else if ("-n".equals(args[i])) {
					if (args.length < i)
						usage();
					++i;
					testsToRun.add(Integer.parseInt(args[i])-1);
				}
			}
			
			if (confSuiteFile == null)
				usage();
			
			// configure log level
			Logger log1 = Logger.getLogger("com.ihr.xbrl.om.formula");
			Logger log2 = Logger.getLogger("com.ihr.xbrl.om.assertions");
			log1.setLevel(Level.toLevel(debugLevel));
			log2.setLevel(Level.toLevel(debugLevel));
			log.setLevel(Level.toLevel(debugLevel));
			
			// Load the conformance suite
			Processor xproc = XMLTools.newProcessor(true);

			XPathCompiler xpc = xproc.newXPathCompiler();
			xpc.declareNamespace("c", CONFORMANCE_NAMESPACE);
			
			DocumentBuilder db = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
			XdmNode xbrl = db.build(new File(confSuiteFile));

			// The output will be sent to a JDom document
			Document page = new Document();
			page.setDocType(new DocType("html","-//W3C//DTD XHTML 1.0 Strict//EN","http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"));
			Element html = new Element("html",XHTMLNS);
			page.addContent(html);
			Element head = new Element("head",XHTMLNS);
			html.addContent(head);
			Element title = new Element("title",XHTMLNS);
			title.setText("Compliance report with the XBRL Formula Conformance Suite");
			head.addContent(title);
			Element style = new Element("style",XHTMLNS);
			head.addContent(style);
			Attribute styleType = new Attribute("type","text/css");
			style.setAttribute(styleType);
			style.setText("td.testOK { background: lightgreen; }\ntd.testFail { background: red; }");			
			Element body = new Element("body",XHTMLNS);
			html.addContent(body);
			Element pageH2 = new Element("h2",XHTMLNS);
			pageH2.setText("Compliance report with the XBRL Formula Conformance suite");
			body.addContent(pageH2);
			Date d = new Date(System.currentTimeMillis());
			SimpleDateFormat dtf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
			Element header3 = new Element("h3",XHTMLNS);
			header3.setText("Date executed: "+dtf.format(d));
			body.addContent(header3);
			Element outTable = new Element("table",XHTMLNS);
			Attribute border1 = new Attribute("border","0");
			outTable.setAttribute(border1);
			body.addContent(outTable);
			
			XPathSelector xps = xpc.compile("for $i in documentation/testcases/testcase return resolve-uri(concat($i/../@root,'/',$i/@uri),base-uri($i))").load();
			xps.setContextItem(xbrl);
			
			XPathExecutable execSelExecDescription = xpc.compile("/c:testcase/c:description");
			XPathExecutable execSelVariation = xpc.compile("/c:testcase/c:variation");
			XPathExecutable execSelId = xpc.compile("@id");
			XPathExecutable execSelDescText = xpc.compile("c:description/text()");
			XPathExecutable execSelExecVariationName = xpc.compile("c:name/text()");
			XPathExecutable execSelFirstFile = xpc.compile("for $i in c:data/*[xs:boolean(@readMeFirst) eq true()] return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelOtherFile = xpc.compile("for $i in c:data/*[(not(@readMeFirst) or false() eq xs:boolean(@readMeFirst)) and not( local-name-from-QName(node-name(.)) eq 'parameter')] return resolve-uri($i,base-uri($i))");
//			XPathExecutable execSelFirstFileAlt = xpc.compile("for $i in c:data/* return resolve-uri($i,base-uri($i))");
			XPathExecutable execSelParamData = xpc.compile("c:data/c:parameter");
			XPathExecutable execSelResult = xpc.compile("c:result/*");
			XPathExecutable execSelResult2 = xpc.compile("if (exists(c:result/@expected)) then true() else false()");
			XPathExecutable execSelResult3 = xpc.compile("if (c:result/@expected eq 'valid') then true() else false()");
//			XPathExecutable execSelError = xpc.compile("c:error");
//			XPathExecutable execSelAssertion = xpc.compile("c:assertionTests");
			XPathExecutable execSelInstance = xpc.compile("doc(resolve-uri(.,base-uri(.)))");
			XPathExecutable execSelCountSatisfied = xpc.compile("string(./@countSatisfied)");
			XPathExecutable execSelCountNotSatisfied = xpc.compile("string(./@countNotSatisfied)");
			XPathExecutable execSelAssertionId = xpc.compile("string(./@assertionID)");
			
			Iterator<XdmItem> iter = xps.iterator();
			long now = System.currentTimeMillis();
			while (iter.hasNext()) {
				XdmItem item = iter.next();
				String strURI = item.getStringValue();
				URI uri = new URI(strURI);
				StreamSource fstc = null;
				try {
					fstc = new StreamSource(uri.toURL().openStream());
				} catch (FileNotFoundException e) {
					log.warn("File: "+uri.toURL().toString()+" cannot be oppened!\nSource document: "+xbrl.getDocumentURI().toString());
					continue; // skip this file
				}
				fstc.setSystemId(strURI);
				DocumentBuilder db2 = XMLTools.newDocumentBuilder(xproc, TreeModel.TINY_TREE, null);
				XdmNode testcases = db2.build(fstc);

				//String description = (String)xp.evaluate("/testcase/@description",testcases,XPathConstants.STRING);
				
//				XPathSelector selExecDescription = xpc.compile("/testcase/@description").load();
//				XPathSelector selVariation = xpc.compile("/testcase/variation").load();
//				XPathSelector selId = xpc.compile("@id").load();
//				XPathSelector selDescText = xpc.compile("description/text()").load();
//				XPathSelector selFirstFile = xpc.compile("for $i in data/*[@readMeFirst = true()] return resolve-uri($i,base-uri($i))").load();
//				XPathSelector selFirstFileAlt = xpc.compile("for $i in data/* return resolve-uri($i,base-uri($i))").load();
//				XPathSelector selResult = xpc.compile("result").load();
				
				XPathSelector selExecDescription = execSelExecDescription.load();
				selExecDescription.setContextItem(testcases);
				String description = selExecDescription.evaluateSingle().getStringValue();
				Element tr01 = new Element("tr",XHTMLNS);
				outTable.addContent(tr01);
				Element td01 = new Element("td",XHTMLNS);
				tr01.addContent(td01);
				td01.setAttribute("style", "background: orange");
				Element link = new Element("a",XHTMLNS);
				link.setText(description);
				link.setAttribute("href", uri.toURL().toString());
				td01.addContent(link);
				//List<?> vars = (List<?>)xp.evaluate("/testcase/variation",testcases,XPathConstants.NODESET);
				
				XPathSelector selVariation = execSelVariation.load();
				selVariation.setContextItem(testcases);
				Iterator<XdmItem> iterVars = selVariation.iterator(); 
				//Iterator<?> iterVars = vars.iterator();
				while (iterVars.hasNext()) {
					Element tr02 = new Element("tr",XHTMLNS);
					outTable.addContent(tr02);
					Element td02 = new Element("td",XHTMLNS);
					tr02.addContent(td02);
					Element table02 = new Element("table",XHTMLNS);
					td02.addContent(table02);
					Attribute at02 = new Attribute("border","1");
					table02.setAttribute(at02);
					Attribute at03 = new Attribute("width","100%");
					table02.setAttribute(at03);
					Element tr03 = new Element("tr",XHTMLNS);
					table02.addContent(tr03);
					XdmItem var = iterVars.next();
					
					XPathSelector selId = execSelId.load();
					selId.setContextItem(var);
					String id = "";
					XdmItem idItem = selId.evaluateSingle();
					if (idItem != null)
						id = idItem.getStringValue();
					//String id = (String)xp.evaluate("@id",var,XPathConstants.STRING);
					
					XPathSelector selDescText = execSelDescText.load();
					selDescText.setContextItem(var);
					String desc = selDescText.evaluateSingle().getStringValue();
					
					XPathSelector selVarNameText = execSelExecVariationName.load();
					selVarNameText.setContextItem(var);
					XdmItem nameNode = selVarNameText.evaluateSingle();
					String name = null;
					if (nameNode != null) 
						name = nameNode.getStringValue();
					
					//String desc = (String)xp.evaluate("description/text()",var,XPathConstants.STRING);
					Element td03 = new Element("td",XHTMLNS);
					tr03.addContent(td03);
					Element href = new Element("a",XHTMLNS);
					href.setAttribute("name", ""+(testNumber+countSkiped+1));
					td03.addContent(href);
					td03.addContent(new Text("Id: "+id+" n:"+(testNumber+countSkiped+1)+(name != null ? (" Name:"+name) : "")));
					Element hrefDown = new Element("a",XHTMLNS);
					hrefDown.setAttribute("href", "#down");
					hrefDown.setText("(dn)");
					td03.addContent(hrefDown);
					Element td04 = new Element("td",XHTMLNS);
					td04.setText("Description: "+desc);
					tr03.addContent(td04);
					Element tr04 = new Element("tr",XHTMLNS);
					table02.addContent(tr04);
					
					// read execution expected result
					boolean assertionOk = false;
					boolean bValidInvalidOk = false;
					QName errorCode = null;
					XPathSelector selResult = execSelResult.load();
					selResult.setContextItem(var);
					XdmValue resultNode = selResult.evaluate();
					testType = UNKNOWN;
					XdmNode rNode = null;
					XdmNode instanceNode = null;
					class ClsAssertionInfo {
						long assertionCountSatisfied = 0;
						long assertionCountNotSatisfied = 0;
						long assertionCountSatisfiedReport = 0;
						long assertionCountNotSatisfiedReport = 0;
						String assertionId = "";
					}
					Vector<ClsAssertionInfo> htAssertions = new Vector<ClsAssertionInfo>();
										
					XPathSelector selResult2 = execSelResult2.load();
					selResult2.setContextItem(var);
					if (selResult2.effectiveBooleanValue()) {
						XPathSelector selResult3 = execSelResult3.load();
						selResult3.setContextItem(var);
						testType |= selResult3.effectiveBooleanValue() ? VALID : INVALID;
					}
					for (XdmItem itemR : resultNode) {
						rNode = (XdmNode)itemR;
						if (rNode.getNodeName().equals(errorQName)) {
							testType |= ERROR_CODE;
							XdmNode errorNode = rNode;
							AtomicSequence qError = errorNode.getUnderlyingNode().atomize();
							String sVal = qError.getStringValue();
							int idx = sVal.indexOf(':');
							String pfx = sVal.substring(0, idx).trim();
							String local = sVal.substring(idx+1).trim();
							String ns = XMLTools.getNamespaceURIforPrefix(pfx, errorNode);
							errorCode = new QName(ns,local,pfx);
							// check if shall replace error code
							QName replaceWith = replaceErrors.get(testNumber+countSkiped+1);
							if (replaceWith != null)
								errorCode = replaceWith;
						} else if (rNode.getNodeName().equals(assertionTestQName)) {
							testType |= ASSERTION;
							XPathSelector selCountSatisfied = execSelCountSatisfied.load();
							XPathSelector selCountNotSatisfied = execSelCountNotSatisfied.load();
							XPathSelector selAssertionId = execSelAssertionId.load();
							selCountSatisfied.setContextItem(rNode);
							selCountNotSatisfied.setContextItem(rNode);
							selAssertionId.setContextItem(rNode);
							
							ClsAssertionInfo asi = new ClsAssertionInfo();
							asi.assertionCountSatisfied = Integer.parseInt( selCountSatisfied.evaluateSingle().getStringValue() );
							asi.assertionCountNotSatisfied = Integer.parseInt( selCountNotSatisfied.evaluateSingle().getStringValue() );
							asi.assertionId = selAssertionId.evaluateSingle().getStringValue();
 							htAssertions.add(asi);
						} else if (rNode.getNodeName().equals(instanceTestQName)) {
							testType |= FORMULA;
							XPathSelector selInstance = execSelInstance.load();
							selInstance.setContextItem(rNode);
							instanceNode = (XdmNode)selInstance.evaluateSingle();
						}
					}

					//String strRes = (String)xp.evaluate("result/@expected",var,XPathConstants.STRING);
					
					XPathSelector selFirstFile = execSelFirstFile.load();
					selFirstFile.setContextItem(var);
					XdmItem selFirstFileResult = selFirstFile.evaluateSingle();
					String varF = null;
					if (selFirstFileResult != null)
						varF = selFirstFileResult.getStringValue();
					XPathSelector selOtherFile = execSelOtherFile.load();
					selOtherFile.setContextItem(var);
					XdmValue selOtherFileResult = selOtherFile.evaluate();
					ArrayList<String> varFs = new ArrayList<>((varF != null & !varF.isEmpty() ? 1 : 0) + selOtherFileResult.size());
					int varFidx = 0;
					if (varF != null && !varF.isEmpty())
						varFs.add(varFidx++, varF);
					for (int idx = 0; idx < selOtherFileResult.size(); ++idx)
						varFs.add(varFidx++, selOtherFileResult.itemAt(idx).getStringValue());
					QName qE = null;
					if ((excludePatterns != null && varF.matches(excludePatterns)) || ((testNumber+countSkiped) < startAt || (testNumber+countSkiped) > endsAt) || (testsToRun.size() > 0 && !testsToRun.contains(testNumber+countSkiped))) {
						Element td05 = new Element("td",XHTMLNS);
						td05.setText("Execution excluded");
						tr04.addContent(td05);
						countSkiped++;
					} else {
						testNumber++;
						debug = (testNumber == 239);
						StringBuffer resultText = new StringBuffer();
						StringBuffer result = new StringBuffer();
						//boolean bError = false;
						boolean instanceEquals = false;
						try {
							if (bShowMemoryUsage) {
								//System.out.println("Total memory at 1:"+Runtime.getRuntime().freeMemory())
								long total = Runtime.getRuntime().totalMemory();
								long free = Runtime.getRuntime().freeMemory();
								log.info("Test number: "+(testNumber+countSkiped)+" Memory at 1: T="+total+" F="+free+" U="+(total-free)+" L=\""+varF+"\"");
							}
							XPathSelector selParameters = execSelParamData.load();
							selParameters.setContextItem(var);
							Properties props = new Properties();
							
							props.setProperty(DTSContainer.PROCESSORS_SEQUENCE, 
									XBRLCoreProcessor.PROCESSORNAME+","+
									XDTProcessor.PROCESSORNAME+","+
									XBRLFormulaProcessor.PROCESSORNAME);
							props.put(DTSContainer.FORMULA_PROCESSOR_PARAMETER_RESOLVER, new ConformanceParametersFactory(selParameters.evaluate()));
							
							/**
							 * IMPORTANT. This flag is set because we pass the conformance suite directly from the
							 * files downloaded from the SVN repository. The conformance suite is using local references
							 * to the official schemas rather than the official copies on the official locations. The
							 * processor cannot handle the same namespace loaded from different URLs and defining the
							 * same data types (this is an XML error). But the processor handles the same namespace
							 * on different URLs when the two schemas "cooperate" on include statements.
							 */
							props.setProperty(DTSContainer.DTS_DISCOVERY_TRUST_NAMESPACES_ONLY_FOR_OFFICIAL_FILES, "true");
							/**
							 * In the formula comformance suite there are files with wrong values in the schemaLocation attribute on
							 * import statements. This directive ignores the error and loads the fine from the official URL which does
							 * not produce any other errors.
							 */
							props.setProperty(DTSContainer.DTS_DISCOVERY_ON_DIFFERENT_FILE_NAMES, DTSContainer.IGNORE_PROPERTY_VALUE);
							/**
							 * The formula specification is not compatible with the latest version of the XBRL 2.1 specification and
							 * the conformance suite test cases contains tests that requires old specification version to be used 
							 */
							props.setProperty(DTSContainer.XBRL_VERSION, DTSContainer.XBRL_VERSION_2_JUL_2008);
							
							props.setProperty(DTSContainer.LOG_LEVEL, debugLevel);
							
							props.setProperty(DTSContainer.FORMULA_PROCESSOR_RETAIN_AFTER_FORMULA_INTANCE_ON_ASSERTION_RESULTS, "true");
//							props.setProperty(DTSContainer.FORMULA_PROCESSOR_RETAIN_INPUT_INTANCE_ON_ASSERTION_RESULTS, "true");
							
							Properties customProperties = testcaseProcessorProperties.get(testNumber+countSkiped+1);
							if (customProperties != null)
								props.putAll(customProperties);
							
							// copy parameter properties
							props.putAll(dtsProps);
							
							DTSContainer dts = DTSContainer.newEmptyContainer(props);
							// Get parameters							
//							ConformanceParameters cfs = new ConformanceParameters(selParameters.evaluate());
							ConformanceFormulasCustomFunctions cff = new ConformanceFormulasCustomFunctions();
							XBRLFormulaProcessor fp = XBRLFormulaProcessor.create(dts);
//							fp.setParametersResolver(cfs);
							fp.addCustomFunctionLibrary(cff);
							URI uriF[] = new URI[varFs.size()];
							int i=0;
							for (String varFu : varFs)
								uriF[i++] = new URI(varFu);
							dts.load(uriF);
							dts.log.setLevel(Level.toLevel(debugLevel));
							XBRLValidationReport report = dts.validate();
							if (errorCode != null) {
								// select from all errors in the validation report
								// the one that matches the errorCode, if any exist
								for (XBRLValidationException ex : GeneralIterable.asIterator(report.getExceptions())) {
									if (ex.getCode() == XBRLValidationException.ENCAPSULATE_XPATH_EXCEPTION) {
										Throwable ex2 = ex.getCause();
										if (ex2 instanceof XFException) {
											XFException xfe = (XFException)ex2;
											if (errorCode.equals( xfe.getQNameError()))
												throw xfe;
										}
										if (ex2 instanceof XFValidationException) {
											XFValidationException xfe = (XFValidationException)ex2;
											if (errorCode.equals( xfe.getQNameError()))
												throw xfe;
										}
									}
									if (errorCode.equals(ex.getQNameError()))
										throw ex;
								}
							}
							report.throwExceptions();
							if (!allowInconsistencies.contains(testNumber+countSkiped))
								report.inconsistenciesAsExceptions();
							
							XBRLValidationReport report2 = null;
							XBRLCoreProcessor proc = XBRLCoreProcessor.create(dts);
							XDTProcessor xdt = XDTProcessor.create(dts);
//							switch (testType) {
							
							// Consider each case individually							
							if (testType == ASSERTION_PLUS_INSTANCE) {
								// validate the returned instance
								if (report.assertionResultsAvailable()) {
									boolean bInsist;
									for (String asId : report.getAssertionResultIDs()) {
										bInsist = false;
										XBRLAssertionResult asr[] = report.getAssertionResult(asId);
										if (asr.length > 0) {
											XBRLInstance instance = asr[0].getAfterFormulaInstance();
											if (instance != null) {
												report2 = proc.validate(asr[0].getAfterFormulaInstance());
												report2.throwExceptions();
												report2.inconsistenciesAsExceptions();
												report2.merge(xdt.validate(asr[0].getAfterFormulaInstance()));
												report2.throwExceptions();
												report2.inconsistenciesAsExceptions();
												instanceEquals = CompareInstancesSemantically.compareInstances(asr[0].getAfterFormulaInstance(), instanceNode, resultText);
											} else
												bInsist = true;
										} else
											bInsist = true;
										if (!bInsist)
											break;
									}
								} else {
									report2 = proc.validate(report.getFormulaProcessorGeneratedInstance());
									report2.throwExceptions();
									report2.inconsistenciesAsExceptions();
									report2.merge(xdt.validate(report.getFormulaProcessorGeneratedInstance()));
									report2.throwExceptions();
									report2.inconsistenciesAsExceptions();
									instanceEquals = CompareInstancesSemantically.compareInstances(report.getFormulaProcessorGeneratedInstance(), instanceNode, resultText);
								}

//								if (instanceEquals) {
								boolean bAllOk = true;
								for (ClsAssertionInfo as : htAssertions) {
									as.assertionCountNotSatisfiedReport = report.getCountAssertionsNotSatisfied(as.assertionId);
									as.assertionCountSatisfiedReport = report.getCountAssertionsSatisfied(as.assertionId);
									if (as.assertionCountSatisfiedReport != as.assertionCountSatisfied || 
											as.assertionCountNotSatisfiedReport != as.assertionCountNotSatisfied) {
										bAllOk = false;
										break; // for
									}
								}
								if (bAllOk || instanceEquals) {
									countOk++;
									assertionOk = true;										
								} else {
									assertionOk = false;
									countError++;										
									errors.add(testNumber+countSkiped);
								}
							} else if (testType == FORMULA) {
								if (report.assertionResultsAvailable()) {
									String asId = report.getAssertionResultIDs().iterator().next();
									XBRLAssertionResult asr[] = report.getAssertionResult(asId);
									if (asr.length > 0) {
										report2 = proc.validate(asr[0].getAfterFormulaInstance());
										report2.throwExceptions();
										report2.inconsistenciesAsExceptions();
										report2.merge(xdt.validate(asr[0].getAfterFormulaInstance()));
										report2.throwExceptions();
										report2.inconsistenciesAsExceptions();
										instanceEquals = CompareInstancesSemantically.compareInstances(asr[0].getAfterFormulaInstance(),instanceNode,resultText);
										if (instanceEquals) {
											countOk++;
										} else {
											countError++;
											errors.add(testNumber+countSkiped);
										}
									} else {
										countError++;
										errors.add(testNumber+countSkiped);
									}
								} else {
									report2 = proc.validate(report.getFormulaProcessorGeneratedInstance());
									report2.throwExceptions();
									report2.inconsistenciesAsExceptions();
									report2.merge(xdt.validate(report.getFormulaProcessorGeneratedInstance()));
									report2.throwExceptions();
									report2.inconsistenciesAsExceptions();
									instanceEquals = CompareInstancesSemantically.compareInstances(report.getFormulaProcessorGeneratedInstance(), instanceNode, resultText);
									if (instanceEquals) {
										countOk++;
									} else {
										countError++;
										errors.add(testNumber+countSkiped);
									}
								}
							} else if ((testType & ERROR_CODE) == ERROR_CODE) {
								if (report.getFormulaProcessorGeneratedInstance() != null) {
									report2 = proc.validate(report.getFormulaProcessorGeneratedInstance());
									report2.throwExceptions();
									report2.inconsistenciesAsExceptions();
									report2.merge(xdt.validate(report.getFormulaProcessorGeneratedInstance()));
									report2.throwExceptions();
									report2.inconsistenciesAsExceptions();
									
									log.trace("Instance document generated when error code was expected");
									
									XBRLInstance cpInstance = null;
									try {
										if (instanceNode != null) {
											cpInstance = new XBRLInstance(report.getFormulaProcessorGeneratedInstance().getContainer(),instanceNode.asSource(),true);
											if (log.isTraceEnabled()) {
												log.trace("Comparing instances:");
												log.trace("conformance suite instance:");
												log.trace(new String(XmlFormater.format(cpInstance.generateDocNode(/*false, null*/)),"UTF-8").toString());
											}
										} else {
											resultText.append("instance node is null maybe this is an aspect test that did not generate an XBRL report");
										}
									} catch (SaxonApiException e) {
										resultText.append("There is an error in the instance document in the conformance suite.\n"+instanceNode.getDocumentURI()+"\n"+e.getMessage());
									} catch (XBRLSerializationException e) {
										resultText.append("There is an error in the instance document in the conformance suite.\n"+instanceNode.getDocumentURI()+"\n"+e.getMessage());
									} catch (NullPointerException e) {
										// TODO:
										resultText.append("There is an error in the instance document in the conformance suite.\n"+e.getMessage());
									}
									
								} else {
									log.trace("No instance document generated. May occur if the processor has validated a taxonomy");
								}
								countError++;
								errors.add(testNumber+countSkiped);
							} else if (testType == ASSERTION) {
								boolean bAllOk = true;
								for (ClsAssertionInfo as : htAssertions) {
									as.assertionCountNotSatisfiedReport = report.getCountAssertionsNotSatisfied(as.assertionId);
									as.assertionCountSatisfiedReport = report.getCountAssertionsSatisfied(as.assertionId);
									if (as.assertionCountSatisfiedReport != as.assertionCountSatisfied || 
											as.assertionCountNotSatisfiedReport != as.assertionCountNotSatisfied) {
										bAllOk = false;
										break; // for
									}
								}
								if (bAllOk) {
									countOk++;
									assertionOk = true;										
								} else {
									assertionOk = false;
									countError++;										
									errors.add(testNumber+countSkiped);
								}
							} else if (testType == VALID) {
								countOk++; // no other errors, so this is OK.
								bValidInvalidOk = true;
							} else if (testType == INVALID) {
								if (report.getFormulaProcessorGeneratedInstance() != null) {
									report2 = proc.validate(report.getFormulaProcessorGeneratedInstance());
									report2.throwExceptions();
									report2.inconsistenciesAsExceptions();
									report2.merge(xdt.validate(report.getFormulaProcessorGeneratedInstance()));
									report2.throwExceptions();
									report2.inconsistenciesAsExceptions();
								}
								countError++; // no error detected so this is an error.
								errors.add(testNumber+countSkiped);
								bValidInvalidOk = false;								
							} else if (testType == UNKNOWN) {
								log.error("Unknown test case type: "+(testNumber+countSkiped));
								instanceEquals = false;
								countError++;
								errors.add(testNumber+countSkiped);
							} else {
								countError++;
								errors.add(testNumber+countSkiped);
								bValidInvalidOk = false;								
							}
							dts.clear();
						} catch (XFValidationException e) {
							qE = e.getQNameError();
							if (errorCode == null || !qE.equals(errorCode)) {
								if (testType == INVALID) {// and no other error is documented
									countOk++;
									bValidInvalidOk = true;
								} else {
									countError++;
									errors.add(testNumber+countSkiped);
								}
							} else {
								countOk++;
								bValidInvalidOk = true;
							}
							resultText.append(" ");
							resultText.append(XBRLValidationException.documentException(e));
							resultText.append("\n");
							result.append(qE.getPrefix());
							result.append(":");
							result.append(qE.getLocalPart());
						} catch (XDTValidationException e) {
							qE = e.getQNameError();
							if (errorCode == null || !qE.equals(errorCode)) {
								if (log.isTraceEnabled())
									e.printStackTrace();
								countError++;
								errors.add(testNumber+countSkiped);
							} else {
								countOk++;
								bValidInvalidOk = true;
							}
							resultText.append(" ");
							resultText.append(XBRLValidationException.documentException(e));
							resultText.append("\n");
							result.append(qE.getPrefix());
							result.append(":");
							result.append(qE.getLocalPart());
						} catch (XBRLValidationException e) {
							qE = e.getQNameError();
							if (errorCode == null || qE == null || !qE.equals(errorCode)) {
								if ((testType & INVALID) == INVALID) {
									countOk++;
									bValidInvalidOk = true;								
								} else {
									countError++;
									errors.add(testNumber+countSkiped);
									bValidInvalidOk = false;								
								}
							} else {
								countOk++;
								bValidInvalidOk = true;								
							}
							resultText.append(" ");
							resultText.append(XBRLValidationException.documentException(e));
							resultText.append("\n");
							if (qE != null) {
								result.append(qE.getPrefix());
								result.append(":");
								result.append(qE.getLocalPart());
							}
						} catch (XFException e) {
							qE = e.getQNameError();
							if (errorCode == null || qE == null || !qE.equals(errorCode)) {
								if ((testType & INVALID) == INVALID) {
									countOk++;
									bValidInvalidOk = true;								
								} else {
									countError++;
									errors.add(testNumber+countSkiped);
									bValidInvalidOk = false;								
								}
							} else {
								countOk++;
								bValidInvalidOk = true;								
							}
							resultText.append(" ");
							resultText.append(XBRLValidationException.documentException(e));
							resultText.append("\n");
							if (qE != null) {
								result.append(qE.getPrefix());
								result.append(":");
								result.append(qE.getLocalPart());
							}
						} catch (XBRLRuntimeException e) {
							countError++;
							errors.add(testNumber+countSkiped);
							resultText.append(" ");
							resultText.append(XBRLValidationException.documentException(e));
							resultText.append("\n");
							result.append(qE == null ? "" : qE.getPrefix());
							result.append(":");
							result.append(qE == null ? XBRLValidationException.documentException(e) : qE.getLocalPart());
						} catch (DTSDiscoveryException e) {
							//qE = e.getQNameError();
							if (errorCode == null /*|| !qE.equals(errorCode)*/) {
								if ((testType & INVALID) == INVALID) {
									countOk++;
									bValidInvalidOk = true;								
								} else {
									countError++;
									errors.add(testNumber+countSkiped);
									bValidInvalidOk = false;								
								}
							} else {
								countOk++;
								bValidInvalidOk = true;								
							}
							resultText.append(" ");
							resultText.append(XBRLValidationException.documentException(e));
							resultText.append("\n");
							result.append(resultText.toString());
//							result.append(qE.getPrefix());
//							result.append(":");
//							result.append(qE.getLocalPart());
						} catch (LicenseVerificationException e) {
							e.printStackTrace();
							System.exit(-1);
						} catch (XBRLInitializationException e) {
							e.printStackTrace();
							System.exit(-1);
						} catch (NullPointerException e) {
							if (log.isTraceEnabled())
								e.printStackTrace();
							resultText.append("NullPointerException received.");
							result.append(resultText.toString());
							countError++;
							errors.add(testNumber+countSkiped);
						} catch (RuntimeException e) {
							if (log.isTraceEnabled())
								e.printStackTrace();
							resultText.append("RuntimeException received.");
							resultText.append(XBRLValidationException.documentException(e));
							result.append(resultText.toString());
							countError++;
							errors.add(testNumber+countSkiped);
						} catch (Throwable e) {
							if (log.isTraceEnabled())
								e.printStackTrace();
							resultText.append("Throwable received.");
							resultText.append(XBRLValidationException.documentException(e));
							result.append(resultText.toString());
							countError++;
							errors.add(testNumber+countSkiped);
						}
						if (bShowMemoryUsage) {
							System.gc();
							long total = Runtime.getRuntime().totalMemory();
							long free = Runtime.getRuntime().freeMemory();
							log.info("Test number: "+(testNumber+countSkiped)+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
						}
						Element td05 = new Element("td",XHTMLNS);
						boolean bTestOk = false;
						String message = null;
//						switch (testType) {
//						case ASSERTION_PLUS_INSTANCE:
							
						if (( testType & ASSERTION_PLUS_INSTANCE) == ASSERTION_PLUS_INSTANCE) {
							bTestOk = assertionOk;
//							message = 							bTestOk ? 
//									"Formula and assertion OK. Expected instance" +
//									" - result is comparison OK "+result.toString()+". "+resultText.toString() :
//									"Formula NOT PASS. Expected instance" + 
//									" - result is "+ (instanceEquals ? "comparison OK" : "comparison Fail") +". "+resultText.toString();
							message = "";
							for (ClsAssertionInfo as : htAssertions) {
//								message = message + (message.length() > 0 ? "<br/>" : ""); 
								message = message +					(bTestOk ? 
									"Assertion OK. Expected "+as.assertionId+ " Count satisfied=" + as.assertionCountSatisfied + " Count not satisfied=" + as.assertionCountNotSatisfied +
									" - result is "+ as.assertionId + " Count satisfied="+as.assertionCountSatisfiedReport + " Count not satisfied="+ as.assertionCountNotSatisfiedReport +". "+resultText.toString() :
									"Assertion NOT PASS. Expected "+as.assertionId+ " Count satisfied=" + as.assertionCountSatisfied + " Count not satisfied=" + as.assertionCountNotSatisfied +
									" - result is "+ as.assertionId + " Count satisfied="+as.assertionCountSatisfiedReport + " Count not satisfied="+ as.assertionCountNotSatisfiedReport +". "+resultText.toString());
							}
//							break;
						}
						if (( testType & FORMULA) == FORMULA) {
//						case FORMULA:
							bTestOk = instanceEquals;
							message = 							bTestOk ? 
									"Formula OK. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) +
									" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "valid") : result.toString()) +". "+resultText.toString() :
									"Formula NOT PASS. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) + 
									" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "comparison Fail") : result.toString()) +". "+resultText.toString();
//							break;
						}
						if (( testType & ERROR_CODE) == ERROR_CODE) {
//						case ERROR_CODE:
							bTestOk = (errorCode.equals(new QName("null")) && qE == null) || errorCode.equals(qE); // errorCode is null if the table of error codes has set null for this error and qE is, for example a DTSCiscoveryError (unexpected and out of sense in this conformance suite, but it happens for test case 575
							message = 							bTestOk ? 
									"Error code OK. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) +
									" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "valid") : result.toString()) +". "+resultText.toString() :
									"Error code NOT PASS. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) + 
									" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "instance generated, no error code detected.") : result.toString()) +". "+resultText.toString();
//							break;
						}
						if (( testType & ASSERTION) == ASSERTION) {
//						case ASSERTION:
							bTestOk = assertionOk;
							message = "";
							for (ClsAssertionInfo as : htAssertions) {
//								message = message + (message.length() > 0 ? "<br/>" : ""); 
								message = message + (bTestOk ? 
									"Assertion OK. Expected "+as.assertionId+ " Count satisfied=" + as.assertionCountSatisfied + " Count not satisfied=" + as.assertionCountNotSatisfied +
									" - result is "+ as.assertionId + " Count satisfied="+as.assertionCountSatisfiedReport + " Count not satisfied="+ as.assertionCountNotSatisfiedReport +". "+resultText.toString() :
									"Assertion NOT PASS. Expected "+as.assertionId+ " Count satisfied=" + as.assertionCountSatisfied + " Count not satisfied=" + as.assertionCountNotSatisfied +
									" - result is "+ as.assertionId + " Count satisfied="+as.assertionCountSatisfiedReport + " Count not satisfied="+ as.assertionCountNotSatisfiedReport +". "+resultText.toString());
							}
//							break;
						}
						if (( testType == UNKNOWN)) {
//						case UNKNOWN:
							bTestOk = false;
							message = "UNKNOWN test type.";
//							break;
						}
						if ((testType & INVALID) == INVALID) {
							bTestOk = bValidInvalidOk;
							if (message == null) {
								message = resultText.length() > 0 ? resultText.toString() : "Test case detected invalid";
							}
						}
						if ((testType & VALID) == VALID) {
							bTestOk = bValidInvalidOk;
							if (message == null) {
								message = resultText.length() > 0 ? resultText.toString() : "Test case is valid";
							}							
						}
						Attribute at06 = new Attribute("class", bTestOk ? "testOK" : "testFail");
						td05.setAttribute(at06);
												
						td05.setText(message);
//							bTestOk ? 
//								"OK. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) +
//								" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "valid") : result.toString()) +". "+resultText.toString() :
//								"NOT PASS. Expected "+ (errorCode == null ? (instanceNode == null ? "valid" : "instance") : errorCode.getPrefix()+":"+errorCode.getLocalPart()) + 
//								" - result is "+ (result.length() == 0 ? (instanceEquals ? "comparison OK" : "comparison Fail") : result.toString()) +". "+resultText.toString());
						tr04.addContent(td05);
						
						Element td06 = new Element("td",XHTMLNS);
						td06.setText(resultText.toString());
						tr04.addContent(td06);
						Element tr05 = new Element("tr",XHTMLNS);
						table02.addContent(tr05);
						Element td07 = new Element("td",XHTMLNS);
						Attribute at07 = new Attribute("colspan","2");
						td07.setAttribute(at07);
						tr05.addContent(td07);
//						td07.setText(varF);
						for (int idx = 0; idx < varFs.size() ; ++idx) {
							if (idx > 0) {
								Element br0 = new Element("br",XHTMLNS);
								td07.addContent(br0);
							}
							String fileUri = varFs.get(idx);
							Element link03 = new Element("a",XHTMLNS);
							link03.setText(fileUri);
	//						link03.setAttribute("href", "xbrl:"+varF.substring(varF.indexOf(':')+1));
							link03.setAttribute("href", fileUri.substring(fileUri.indexOf(':')+1));
							td07.addContent(link03);
						}
					}
				} // Shall we close stream here?
//				if (bShowMemoryUsage) {
//					//System.runFinalization();
//					System.gc();
//					long total = Runtime.getRuntime().totalMemory();
//					long free = Runtime.getRuntime().freeMemory();
//					System.out.println("Test number: "+testNumber+" "+"Memory at 2: T="+total+" F="+free+" U="+(total-free));
//				}
			}
			Element table03 = new Element("table",XHTMLNS);
			body.addContent(table03);
			Element tr05 = new Element("tr",XHTMLNS);
			table03.addContent(tr05);
			Element td07 = new Element("td",XHTMLNS);
			tr05.addContent(td07);
			td07.setText("Not passed:");
			Element td08 = new Element("td",XHTMLNS);
			tr05.addContent(td08);
			Element linkDownPair = new Element("a",XHTMLNS);
			linkDownPair.setAttribute("name", "down");
			td08.addContent(linkDownPair);
			td08.addContent(new Text((new Integer(countError)).toString()));
			td08.addContent(getErrorsCollection(errors));
			Element tr06 = new Element("tr",XHTMLNS);
			table03.addContent(tr06);
			Element td09 = new Element("td",XHTMLNS);
			tr06.addContent(td09);
			td09.setText("Passed:");
			Element td10 = new Element("td",XHTMLNS);
			tr06.addContent(td10);
			td10.setText((new Integer(countOk)).toString());
			Element tr07 = new Element("tr",XHTMLNS);
			table03.addContent(tr07);
			Element td11 = new Element("td",XHTMLNS);
			tr07.addContent(td11);
			td11.setText("Not executed:");
			Element td12 = new Element("td",XHTMLNS);
			tr07.addContent(td12);
			td12.setText((new Integer(countSkiped)).toString());

			// produce the output to stdout or file
			Format ft = Format.getPrettyFormat();
			ft.setEncoding("ISO-8859-1");
			XMLOutputter outputter = new XMLOutputter(ft);
			String s = outputter.outputString(page);
			if (outputFile != null) {
				File f = new File(outputFile);
				FileWriter fw = new FileWriter(f);
				fw.write(s);
				fw.flush();
				fw.close();
			} else {
				System.out.println(s);
			}
			long now2 = System.currentTimeMillis();
			log.info("Total execution time "+(now2-now)/1000+"s");
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (MalformedURLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (SaxonApiException e) {
			e.printStackTrace();
		} catch (XPathException e) {
			e.printStackTrace();
		}
		System.exit(0);
	}

	/**
	 * @param errors
	 * @return
	 */
	private static Collection<Content> getErrorsCollection(LinkedList<Integer> errors) {
		ArrayList<Content> al = new ArrayList<>(errors.size());
		for (Integer error : errors) {
			al.add(new Text(" "));
			Element link = new Element("a",XHTMLNS);
			link.setAttribute("href", "#"+error);
			link.setText(""+error);
			al.add(link);
		}
		return al;
	}

}
