/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 05/08/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;

import javax.xml.namespace.QName;

import net.sf.saxon.value.QNameValue;
import net.sf.saxon.value.StringValue;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XBRLAttribute;
import com.ihr.xbrl.om.XMLFragment;
import com.ihr.xbrl.om.exLinks.GenericLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.exLinks.XBRLResource;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLImport;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;

/**
 * This example creates a taxonomy and a formula linkbase that is a clone of
 * the following example in the formula conformance suite:
 * 
 * http://publicsvn.xbrl.org/svn/public/formula-conformance/2009-06-22/tests/10000%20Formula/14020-Formula-UseCases-Movement/14020-simpleMovement-concepts.xsd
 * 
 * Note the syntax of the formula linkbase is based on the generic linkbase.
 * All resources are generic resources. All relationships are generic
 * relationships. The formula specification already provides schemas
 * where role URIs and arcrole URIs are defined. This sample application
 * assumes the official schemas are used so running this example
 * requires Internet connection.
 * 
 * @author Ignacio
 *
 */
public class SampleCreateFormulaLinkbase {

	/**
	 * @param args
	 * @throws Exception In case of any errors
	 */
	public static void main(String[] args) throws Exception {
		DTSContainer dts = DTSContainer.newEmptyContainer();

		// load official schemas for the formula linkbase
		URI uris[] = new URI[] {
				new URI("http://www.xbrl.org/2008/generic-link.xsd"),
				new URI("http://www.xbrl.org/2008/formula.xsd"),
				new URI("http://www.xbrl.org/2008/variable.xsd"),
				new URI("http://www.xbrl.org/2008/concept-filter.xsd"),
				new URI("http://www.xbrl.org/2008/period-filter.xsd"),
				new URI("http://www.xbrl.org/2008/generic-label.xsd")
		};
		dts.load(uris);
		
		XBRLTaxonomy tx = new XBRLTaxonomy(dts, true);
		
		String targetNamespace = "http://xbrl.org/formula/conformance/example";
		tx.setURI(new URI("14020-simpleMovement-concepts.xsd"));
		tx.setTargetNamespace(targetNamespace);
		tx.getDocumentRootNode().addNamespace("eg", targetNamespace);
		
		// This is required when new documents are created programatically because the load method
		// automatically adds all discovered documents to the DTS but new documents created are not
		// added automatically. Documents are indexed in the DTS by the document key. The document key
		// is build from the document URI. so call this method after setting a URI to the new document.
		dts.addDocumentToDTS(tx);
		
		// Import the XBRL Instance schema
		XBRLImport.make(tx, XBRL.XBRL_instanceNS, XBRL.XBRL_instanceURL);
//		tx.addImport(new XBRLImport(tx, XBRL.XBRL_instanceNS, XBRL.XBRL_instanceURL));
		
		XBRLXSDTypeDefinition monetaryType = 
			new XBRLXSDTypeDefinition(
					dts,
					new QName(XBRL.XBRL_instanceNS,"monetaryItemType")
			);
		
		QName itemSg = new QName(XBRL.XBRL_instanceNS,"item");
		
		XBRLItem balance = new XBRLItem(
				tx,
				"balance",
				null,
				monetaryType,
				itemSg,
				XBRLItem.PeriodType.INSTANT,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);
		
		XBRLItem changes = new XBRLItem(
				tx,
				"changes",
				null,
				monetaryType,
				itemSg,
				XBRLItem.PeriodType.DURATION,
				XBRLItem.BalanceType.NO_BALANCE,
				false,
				true,
				XMLElementDefinition.UNDEFINED,
				XMLElementDefinition.UNDEFINED,
				null);
		
		// Create the linkbase container and attach it to the taxonomy
		XBRLLinkbase linkBase = new XBRLLinkbase(dts, true);
		// we need a role type for any extended link. Get the role defined in the generic linkbase
		XBRLRoleType roleLink = dts.getRoleType("http://www.xbrl.org/2008/role/link").firstElement();
		// Create a new presentation linkbase in the DTS
		GenericLinkbase lb = new GenericLinkbase(linkBase, roleLink);
		
		// put a uri for the new labels linkbase
		linkBase.setURI(new URI("14020-simpleMovement-formula.xml"));
		dts.addDocumentToDTS(linkBase);
		// add the linkbase to the taxonomy
		tx.addLinkbase(linkBase);

		
		// Create a formula:formula resource
		XBRLResource formula = new XBRLResource(lb, new QName("http://xbrl.org/2008/formula","formula","formula"),true);
		formula.addAttribute(new QName("value"), "$v:beginningBalance + $v:changes");
		formula.addAttribute(new QName("source"), "v:beginningBalance");
		formula.addAttribute(new QName("aspectModel"), "dimensional");
		formula.addAttribute(new QName("implicitFiltering"), "true");
		XMLFragment fDecimals = new XMLFragment(dts);
		fDecimals.setNodeName(new QName("http://xbrl.org/2008/formula","decimals","formula"));
		fDecimals.setValue(new StringValue("0"));
		formula.addChild(fDecimals);
		XMLFragment fAspects = new XMLFragment(dts);
		fAspects.setNodeName(new QName("http://xbrl.org/2008/formula","aspects","formula"));
		formula.addChild(fAspects);
		XMLFragment fPeriod = new XMLFragment(dts);
		fPeriod.setNodeName(new QName("http://xbrl.org/2008/formula","period","formula"));
		fAspects.addChild(fPeriod);
		XMLFragment fInstant = new XMLFragment(dts);
		fInstant.setNodeName(new QName("http://xbrl.org/2008/formula","instant","formula"));
		fInstant.addAttribute(new QName("value"), "xfi:period-end(xfi:period($v:changes))");
		fPeriod.addChild(fInstant);
		
		// Create the variable:factVariable resources
		XBRLResource variable_beginningBalance = new XBRLResource(lb, new QName("http://xbrl.org/2008/variable", "factVariable", "variable"),true);
		variable_beginningBalance.addAttribute(new QName("bindAsSequence"), "false");

		XBRLResource variable_changes = new XBRLResource(lb, new QName("http://xbrl.org/2008/variable","factVariable","variable"),true);
		variable_changes.addAttribute(new QName("bindAsSequence"), "false");
		
		// Create the concept filter resource
		XBRLResource filter_balance = new XBRLResource(lb,new QName("http://xbrl.org/2008/filter/concept","conceptName","cf"),true);
		XMLFragment fConcept1 = new XMLFragment(dts);
		fConcept1.setNodeName(new QName("http://xbrl.org/2008/filter/concept","concept","cf"));
		filter_balance.addChild(fConcept1);
		XMLFragment fQName1 = new XMLFragment(dts);
		fQName1.setNodeName(new QName("http://xbrl.org/2008/filter/concept","qname","cf"));
		fQName1.setValue(new QNameValue("c",balance.getQName().getNamespaceURI(), balance.getQName().getLocalPart()));
		fConcept1.addChild(fQName1);
		
		XBRLResource filter_changes = new XBRLResource(lb,new QName("http://xbrl.org/2008/filter/concept","conceptName","cf"),true);
		XMLFragment fConcept2 = new XMLFragment(dts);
		fConcept2.setNodeName(new QName("http://xbrl.org/2008/filter/concept","concept","cf"));
		filter_changes.addChild(fConcept2);
		XMLFragment fQName2 = new XMLFragment(dts);
		fQName2.setNodeName(new QName("http://xbrl.org/2008/filter/concept","qname","cf"));
		fQName2.setValue(new QNameValue("c",changes.getQName().getNamespaceURI(), changes.getQName().getLocalPart()));
		fConcept2.addChild(fQName2);
		
		// Create label resources
		// Obtain the role for labels Note this is NOT the standard role. This is defined in the gLabelTx schema.
		XBRLRoleType labelRole = dts.getRoleType("http://www.xbrl.org/2008/role/label").firstElement();
		
		XBRLResource formula_label = new XBRLResource(lb,new QName("http://xbrl.org/2008/label","label","label"),true);
		formula_label.setRoleType(labelRole);
		formula_label.setLang("en");
		formula_label.setValue(new StringValue("Computes ending balance from beginning balance plus changes."));
		
		XBRLResource variable_beginningBalance_label = new XBRLResource(lb,new QName("http://xbrl.org/2008/label","label","label"),true);
		variable_beginningBalance_label.setRoleType(labelRole);
		variable_beginningBalance_label.setLang("en");
		variable_beginningBalance_label.setValue(new StringValue("Binds to any beginning balance"));
		
		XBRLResource variable_changes_label = new XBRLResource(lb,new QName("http://xbrl.org/2008/label","label","label"),true);
		variable_changes_label.setRoleType(labelRole);
		variable_changes_label.setLang("en");
		variable_changes_label.setValue(new StringValue("Binds to any changes with a duration matching the beginning balance instant"));
		
		// create the instance duration period filter.
		XBRLResource periodFilter = new XBRLResource(lb, new QName("http://xbrl.org/2008/filter/period","instantDuration","pf"),true);
		periodFilter.addAttribute(new QName("variable"), "v:changes");
		periodFilter.addAttribute(new QName("boundary"), "start");
		
		// Start creating relationships
		XBRLArcroleType element_label = dts.getArcroleType("http://xbrl.org/arcrole/2008/element-label").firstElement();
		
		XBRLRelationship rel1 = new XBRLRelationship(lb,GenericLinkbase.standard_arc,formula,formula_label,element_label,null,true);
		rel1.setOrder(1.0);
		
		XBRLRelationship rel2 = new XBRLRelationship(lb,GenericLinkbase.standard_arc,variable_beginningBalance,variable_beginningBalance_label,element_label,null,true);
		rel2.setOrder(1.0);
		
		XBRLRelationship rel3 = new XBRLRelationship(lb,GenericLinkbase.standard_arc,variable_changes,variable_changes_label,element_label,null,true);
		rel3.setOrder(1.0);
		
		XBRLArcroleType variable_set = dts.getArcroleType("http://xbrl.org/arcrole/2008/variable-set").firstElement();
		
		XBRLRelationship rel4 = new XBRLRelationship(lb,new QName("http://xbrl.org/2008/variable","variableArc","variable"),formula,variable_changes,variable_set,new XBRLAttribute[] { new XBRLAttribute(new QName("name"), "v:changes") },true);
		rel4.setOrder(1.0);
		
		XBRLRelationship rel5 = new XBRLRelationship(lb,new QName("http://xbrl.org/2008/variable","variableArc","variable"),formula,variable_beginningBalance,variable_set,new XBRLAttribute[] { new XBRLAttribute(new QName("name"), "v:beginningBalance") },true);
		rel5.setOrder(2.0);

		XBRLArcroleType variable_filter = dts.getArcroleType("http://xbrl.org/arcrole/2008/variable-filter").firstElement();
		
		XBRLRelationship rel6 = new XBRLRelationship(lb,new QName("http://xbrl.org/2008/variable","variableFilterArc","variable"),variable_changes,filter_changes,variable_filter,new XBRLAttribute[] { new XBRLAttribute(new QName("complement"), "false"), new XBRLAttribute(new QName("cover"), "true") },true);
		rel6.setOrder(1.0);
		
		XBRLRelationship rel7 = new XBRLRelationship(lb,new QName("http://xbrl.org/2008/variable","variableFilterArc","variable"),variable_beginningBalance,filter_balance,variable_filter,new XBRLAttribute[] { new XBRLAttribute(new QName("complement"), "false"), new XBRLAttribute(new QName("cover"), "true") },true);
		rel7.setOrder(1.0);
		
		XBRLRelationship rel8 = new XBRLRelationship(lb,new QName("http://xbrl.org/2008/variable","variableFilterArc","variable"),variable_beginningBalance,periodFilter,variable_filter,new XBRLAttribute[] { new XBRLAttribute(new QName("complement"), "false"), new XBRLAttribute(new QName("cover"), "true") },true);
		rel8.setOrder(1.0);

		dts.save(/*true*/);
		
		System.exit(0);
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
