/*
 *  2009 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2009 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 25/05/2009
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.net.URI;
import java.net.URISyntaxException;

import javax.xml.namespace.QName;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XBRLAttribute;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLInconsistentDTSException;
import com.ihr.xbrl.om.errors.XBRLSerializationException;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.exLinks.DefinitionLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLExtendedLink;
import com.ihr.xbrl.om.exLinks.XBRLLinkbase;
import com.ihr.xbrl.om.exLinks.XBRLRelationship;
import com.ihr.xbrl.om.taxonomy.XBRLArcroleType;
import com.ihr.xbrl.om.taxonomy.XBRLImport;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.taxonomy.XBRLXSDTypeDefinition;
import com.ihr.xbrl.om.taxonomy.XMLElementDefinition;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.rs.lic.LicenseVerificationException;

/**
 * In this example we will create a taxonomy with one element "Sales" then
 * a dimension "ByRegionDimension" and three dimension members "North", "Center", "South"
 * The dimension will be linked with the primary item using an hypercube 
 * "SalesByRegion"
 * <p>
 * The example shows how to define the concepts in the DTS and the required
 * relationships in order to set-up the "dimensional space" of the concept
 * definition "Sales"
 * <p>
 * This demo is using release 2.6.4 of the XBRL API. Some functions may not
 * exist on previous releases.
 * 
 * @author Ignacio
 *
 */
public class SampleCreateTaxonomyWithDimensions {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			DTSContainer dts = DTSContainer.newEmptyContainer();
			
			// Load the dimensions schema to the DTS
			dts.load(new URI(XBRL.XBRL_dimension_t_URL));
			
			// Create a new taxonomy
			XBRLTaxonomy tx = new XBRLTaxonomy(dts, true);
			
			String targetNamespace = "http://www.example.com/newTaxonomyWithDimensions";
			tx.setURI(new URI("taxonomyWithDimensions.xsd"));
			tx.setTargetNamespace(targetNamespace);
			tx.getDocumentRootNode().addNamespace("tx", targetNamespace);
			
			// Import the XBRL Instance schema
			tx.addImport(new XBRLImport(tx,XBRL.XBRL_instanceNS,XBRL.XBRL_instanceURL));
			// Import the dimensions taxonomy
			tx.addImport(new XBRLImport(tx,XBRL.XBRL_dimension_t_NS,XBRL.XBRL_dimension_t_URL));
			
			// This is required when new documents are created programatically because the load method
			// automatically adds all discovered documents to the DTS but new documents created are not
			// added automatically. Documents are indexed in the DTS by the document key. The document key
			// is build from the document URI. so call this method after setting a URI to the new document.
			dts.addDocumentToDTS(tx);
			
			XBRLXSDTypeDefinition monetaryType = dts.getTypeDefinition(new QName(XBRL.XBRL_instanceNS,"monetaryItemType"));
			XBRLXSDTypeDefinition stringType = dts.getTypeDefinition(new QName(XBRL.XBRL_instanceNS,"stringItemType"));

			// Sales concept (the primary item)
			XBRLItem sales = new XBRLItem(
					tx,
					"Sales",
					null,
					monetaryType,
					XBRLItem.substitutionGroup,
					XBRLItem.PeriodType.DURATION,
					XBRLItem.BalanceType.NO_BALANCE,
					false,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);

			// Define the ByRegionDimension
			XBRLItem byRegionDimension = new XBRLItem(
					tx,
					"ByRegionDimension",
					null,
					stringType,
					XDTProcessor.dimensionItemQName,
					XBRLItem.PeriodType.DURATION,
					XBRLItem.BalanceType.NO_BALANCE,
					true,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);
			
			// Define the North domain member
			XBRLItem north = new XBRLItem(
					tx,
					"North",
					null,
					stringType,
					XBRLItem.substitutionGroup,
					XBRLItem.PeriodType.DURATION,
					XBRLItem.BalanceType.NO_BALANCE,
					true,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);
			
			// Define the Center domain member
			XBRLItem center = new XBRLItem(
					tx,
					"Center",
					null,
					stringType,
					XBRLItem.substitutionGroup,
					XBRLItem.PeriodType.DURATION,
					XBRLItem.BalanceType.NO_BALANCE,
					true,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);
			
			// Define the South domain member
			XBRLItem south = new XBRLItem(
					tx,
					"South",
					null,
					stringType,
					XBRLItem.substitutionGroup,
					XBRLItem.PeriodType.DURATION,
					XBRLItem.BalanceType.NO_BALANCE,
					true,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);
			
			// Define the SalesByRegion
			XBRLItem salesByRegion = new XBRLItem(
					tx,
					"SalesByRegionHypercume",
					null,
					stringType,
					XDTProcessor.hypercubeItemQName,
					XBRLItem.PeriodType.DURATION,
					XBRLItem.BalanceType.NO_BALANCE,
					true,
					true,
					XMLElementDefinition.UNDEFINED,
					XMLElementDefinition.UNDEFINED,
					null);
			
			// Now create the required relationships in the definition linkbase
			
			// Create the linkbase container and attach it to the taxonomy
			XBRLLinkbase linkBase = new XBRLLinkbase(dts, true);
			// we need a role type for any extended link. Get the standard role from the DTS
			XBRLRoleType roleLink = dts.getRoleType(XBRLExtendedLink.standard_role_URI).firstElement();
			// Create a new presentation linkbase in the DTS
			DefinitionLinkbase lb = new DefinitionLinkbase(linkBase, roleLink);
						
			XBRLArcroleType all = dts.getArcroleType(XDTProcessor.all_arcrole_URI).firstElement();
			
			// Create the all relationship form the Sales concept to the hypercube
			XBRLRelationship.make(
					lb,
					DefinitionLinkbase.standard_arc,
					sales,
					salesByRegion,
					all,
					new XBRLAttribute[] {
						new XBRLAttribute(new QName(XBRL.XBRL_dimension_t_NS,"closed"), "false"),
						new XBRLAttribute(new QName(XBRL.XBRL_dimension_t_NS,"contextElement"), "segment")
					},
					true
			);
						
			XBRLArcroleType hyp_dim = dts.getArcroleType(XDTProcessor.hypercube_dimension_arcrole_URI).firstElement();
			
			// Link the dimension with the hypercube
			XBRLRelationship.make(
					lb,
					DefinitionLinkbase.standard_arc,
					salesByRegion,
					byRegionDimension,
					hyp_dim,
					null,
					true
			);
			
			XBRLArcroleType dim_dom = dts.getArcroleType(XDTProcessor.dimension_domain_arcrole_URI).firstElement();
			
			// Link the dimension with the hypercube
			XBRLRelationship.make(
					lb,
					DefinitionLinkbase.standard_arc,
					byRegionDimension,
					north,
					dim_dom,
					null,
					true
			);
			
			// Link the dimension with the hypercube
			XBRLRelationship.make(
					lb,
					DefinitionLinkbase.standard_arc,
					byRegionDimension,
					center,
					dim_dom,
					null,
					true
			);
			
			// Link the dimension with the hypercube
			XBRLRelationship.make(
					lb,
					DefinitionLinkbase.standard_arc,
					byRegionDimension,
					south,
					dim_dom,
					null,
					true
			);
			
			// put a uri for the new labels linkbase
			linkBase.setURI(new URI("definition.xml"));
			dts.addDocumentToDTS(linkBase);
			            
            // Save the whole DTS
            dts.save(/*true*/); // in 2.6.4
			//tx.save(true);
			//linkBase.save(true);
            
			System.exit(0);
		} catch (DTSDiscoveryException e) {
			e.printStackTrace();
		} catch (URISyntaxException e) {
			e.printStackTrace();
		} catch (XBRLValidationException e) {
			e.printStackTrace();
		} catch (XBRLInconsistentDTSException e) {
			e.printStackTrace();
		} catch (XBRLSerializationException e) {
			e.printStackTrace();
		} catch (LicenseVerificationException e) {
			e.printStackTrace();
		}
	}

}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
