/*
 *  2014 by Ignacio Hernandez-Ros.
 *
 * This file is property of Reporting Standard S.L. and shall not be used outside the company unless you
 * have written permissions for different purposes.
 *
 * If you have received this file from Reporting Standard, you may have the following permissions:
 *
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2014 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 20/04/2014
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.util.ArrayList;
import java.util.Properties;

import javax.xml.namespace.QName;

import com.ihr.xbrl.dts.CompanyKey;
import com.ihr.xbrl.dts.DDimension;
import com.ihr.xbrl.dts.DefaultUnitProviderInterface;
import com.ihr.xbrl.dts.Dimension;
import com.ihr.xbrl.dts.DimensionalInstanceExplorer;
import com.ihr.xbrl.dts.HasAccessSelectedRole;
import com.ihr.xbrl.dts.HasDimensionsTableHeaderItem;
import com.ihr.xbrl.dts.IdentifierDimension;
import com.ihr.xbrl.dts.InstanceExplorerEvent;
import com.ihr.xbrl.dts.InstanceExplorerListener;
import com.ihr.xbrl.dts.LangsProvider;
import com.ihr.xbrl.dts.LangsProviderInstance;
import com.ihr.xbrl.dts.Member;
import com.ihr.xbrl.dts.PeriodDimensionInstant;
import com.ihr.xbrl.dts.XDimension;
import com.ihr.xbrl.gui.interfaces.HasBoySupport;
import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRL;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.errors.DTSDiscoveryException;
import com.ihr.xbrl.om.errors.XBRLInitializationException;
import com.ihr.xbrl.om.instance.XBRLFactItem;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.taxonomy.XBRLItem;
import com.ihr.xbrl.om.taxonomy.XBRLRoleType;
import com.ihr.xbrl.om.xdt.XDTInfosetConsumer;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.ihr.xbrl.util.GeneralIterable;
import com.rs.lic.LicenseVerificationException;

/**
 * This is just a test case to show how to use the DimensionalInstanceExplorer to
 * extract values from an instance document using queries.
 * 
 * This example could be used to define a process to query the content of an XBRL
 * report.
 * 
 * a screen capture with the table representation is on the file named: "GBP_TableView.jpg"
 * 
 * @author Ignacio
 *
 */
public class SampleDimensionalInstanceExplorer {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			if (args.length == 0) {
				args = new String[1];
				args[0] = "GPB_NewDate.xbrl";
			}
			Properties props = new Properties();
			
			// configure the processors sequence. We need to perform Dimensions
			// validation and XBRL 2.1 validation
			props.setProperty(DTSContainer.PROCESSORS_SEQUENCE, 
				XBRLCoreProcessor.PROCESSORNAME+","+
				XDTProcessor.PROCESSORNAME
			);
			
			DTSContainer dts = DTSContainer.newEmptyContainer(props);
			XBRLInstance instance = (XBRLInstance)dts.load(new File(args[0]).toURI());
			
			// validate the loaded document so all processors are created and linked
			// to the DTSContainer. We shall access them later
			dts.validate();
			
			System.out.println("Instalce loaded");
			
			// obtain the default langs provider
			LangsProvider lp = LangsProviderInstance.getLangsProviderInstance();
			
			// There is no a default implementation of the HasBoySupport in the API
			// so we need to create our own. In our case, December 31st will be the
			// closing books date for the company
			HasBoySupport hb = new HasBoySupport() {
	
				@Override
				public int askUserNewEoyMonth() {
					// there is no need to implement this method
					return 11;
				}

				@Override
				public int getEoyMonth() {
					return 11; // december
				}

				@Override
				public void setEoyMonth(int newEoyMonth) {
					// there is no need to implement this method
				}

				@Override
				public int getBoyMonth() {
					// there is no need to implement this method
					return 0;
				}
			};
			
			// We need to provide an instance of the HasDimensionsTableHeaderItem to the
			// constructor of the DimensionalInstanceExplorer. This class shall return
			// the primary item that is "selected" (either in a row, column or Z axis)
			// and for which dimensions are being explored.
			// our implementation searches for values of one specific primary item so
			// we will just return it here
			final XBRLItem primaryItem = (XBRLItem)dts.getConcept(new QName("http://www.xbrl.in/rbi/gpb/2009-02-06", "ForeignCurrencyMaturityMismatch"));
			
			HasDimensionsTableHeaderItem hdth = new HasDimensionsTableHeaderItem() {
				
				@Override
				public XBRLItem getSelectedPrimaryItem() {
					return primaryItem;
				}
			};
			
			// We need to provide an instance of the HasAccessSelectedRole interface that
			// shall return the application selected role type where dimensions will be
			// search on
			// our implementation returns just one role we are searching values on		
			final XBRLRoleType role = dts.getRoleType("http://www.xbrl.in/rbi/gpb/2009-02-06/role/GapsPositionandBalances").firstElement();
			
			HasAccessSelectedRole hdsr = new HasAccessSelectedRole() {
				
				@Override
				public XBRLRoleType getSelectedRole() {
					return role;
				}
			};
			
			// We need to provide an instance of the DefaultUnitProviderInterface interface
			// that will be responsible of returning the default unit the user works with
			// in our case we will return the unit used in the sample instance
			final QName defaultUnit = new QName(XBRL.XBRL_iso4217NS,"INR");
			
			DefaultUnitProviderInterface unitProvider = new DefaultUnitProviderInterface() {
				
				@Override
				public QName getDefaultUnit() {
					return defaultUnit;
				}
			};
			
//			// We need to provide an instance of the HandleInstantMode interface that is a class
//			// that returns the user preference about how to match instance data inside a period
//			// the possible values are defined in the IntantMode enumeration. In our case this
//			// information will not be used but in the most common cases this is that the instanc
//			// shall match the period end so we will just return that constant value
//			HandleInstantMode him = new HandleInstantMode() {
//				
//				@Override
//				public InstantMode getInstantMode() {
//					return InstantMode.EOP;
//				}
//			};
//			
			// as data will be collected in an event handler we need
			// to use a "final" object to wrap all information collected
			class ClassDataCollected {
				DDimension dim1;
				XDimension dim2;
				IdentifierDimension dim3;
				PeriodDimensionInstant dim4;
			}
			
			final ClassDataCollected data = new ClassDataCollected();
			
			// before we create the DimensionalInstanceExplorer instance, we are going to create
			// an instance of the InstanceExplorerListener and register the listener with the
			// DimensionalInstanceExplorer. The DimensionalInstanceExplorer will send messages
			// to the listener when the instance is processed and we will capture the events
			// to catch the values we will be later searching for
			DimensionalInstanceExplorer.addInstanceExplorerListener(new InstanceExplorerListener() {
	
				@Override
				public void instanceEvent(InstanceExplorerEvent e) {
					
					// put a break point here and execute the application in debug mode
					System.out.println("InstanceExplorerEvent event received: "+e.getEvent());
	
					// this is the list of possible event codes. Event parameters depends on event code
					/*
						COLUMN_ADDED = 100;
						COLUMN_DELETED = 101;
						FACT_ADDED = 200;
						FACT_DELETED = 201;
						COMPANY_ADDED = 300;
						COMPANY_DELETED = 301;
						FACT_CHANGED_CONTEXT = 400;
						START_INSTANCE = 600;
						END_INSTANCE = 601;
						START_ALLFACTS = 602;
						END_ALLFACTS = 603;
						DIMENSION_ADDED = 500;
						DIMENSION_DELETED = 501;
						DIMENSIONS_FINISHED = 502;
						MEMBER_ADDED = 503;				 
					 */
					if (e.getEvent() == InstanceExplorerEvent.START_INSTANCE) {
						// this is the first event
						System.out.println("Start collecting data from instance");
					} else if (e.getEvent() == InstanceExplorerEvent.COMPANY_ADDED) {
						CompanyKey ck = (CompanyKey)e.getObject();
						System.out.println("Comany added "+ck.getScheme()+" "+ck.getIdentifier());
					} else if (e.getEvent() == InstanceExplorerEvent.COLUMN_ADDED) {						
						System.out.println("Data column "+e.getObject().toString());
					} else if (e.getEvent() == InstanceExplorerEvent.START_ALLFACTS) {
						System.out.println("Start processing facts in the report");
					} else if (e.getEvent() == InstanceExplorerEvent.DIMENSION_ADDED) {
						// obtain the dimension just added
						Dimension<?> dim = (Dimension<?>)e.getObject();						
						System.out.println("There is a new dimension discovered: Type: "+dim.getClass().getCanonicalName());
						if (data.dim1 == null && dim instanceof DDimension) {
							// this is a dimension coming from a parent-child set of relationships
							data.dim1 = (DDimension)dim;
						} else if (data.dim2 == null && dim instanceof XDimension) {
							// this is a dimension coming from a definition linkbase
							data.dim2 = (XDimension)dim;
						} else if (data.dim3 == null && dim instanceof IdentifierDimension) {
							// this is a dimension coming from the context of the reported fact. we call it a "virtual" dimension
							data.dim3 = (IdentifierDimension)dim;
						} else if (data.dim4 == null && dim instanceof PeriodDimensionInstant) {
							// this is a dimension coming from the context of the reported fact. we call it also a "virtual" dimension
							data.dim4 = (PeriodDimensionInstant)dim;
						}
					} else if (e.getEvent() == InstanceExplorerEvent.MEMBER_ADDED) {
						System.out.println("There is a new dimension member discovered: "+e.getObject().toString());
					} else if (e.getEvent() == InstanceExplorerEvent.END_ALLFACTS) {
						System.out.println("End of processing facts");
					} else if (e.getEvent() == InstanceExplorerEvent.END_INSTANCE) {
						System.out.println("End of processing instance");
					}					
				}
				
			});
			
			// Now we just create the instance of the DimensionalInstanceExplorer
			DimensionalInstanceExplorer di = new DimensionalInstanceExplorer(lp, hb, hdth, hdsr, unitProvider/*, him*/);
			
			// we need to configure the DimensionalInstanceExplorer and provide the XDTProcessor
			XDTProcessor xdt = dts.getProcessor(XDTProcessor.PROCESSORNAME);
			XDTInfosetConsumer xdtInfo = new XDTInfosetConsumer(xdt);
			di.setXDTProcessors(xdt, xdtInfo);
			
			// This shall start launching events
			di.setInstance(instance);
			
			// in order to search for a specific fact in the instance using a search criteria,
			// we shall call the findFact method with an array list of Member<?> instances to search for
			// so let's create the array and add some members now
			
			ArrayList<Member<?>> allMembers = new ArrayList<Member<?>>();
			
			// We previously collected the dimensions that exists for all facts in the selected role
			// so we can now explore the dimensions and pick up the members we want to search facts for
			
			// from the parent-child "virtual" dimension, we will pick up the first element
			allMembers.add( data.dim1.getDomain().iterator().next() );
			
			// from the "virtual" dimension used to select the company identifiers, we will use the first element 
			allMembers.add( data.dim3.getDomain().iterator().next() );
			
			// from the "virtual" dimension used to select the time, we will use the first element 
			allMembers.add( data.dim4.getDomain().iterator().next() );
			
			// from the "real" dimension defined in the definition linkbase, we will use one of the leaf members
			// in the tree. The root node has no facts reported
			Member<?> firstNode = data.dim2.getMembers().iterator().next();
			for (Member<?> everyColumn : GeneralIterable.asIterator(firstNode.getDomain().iterator())) {
				Member<?> one = data.dim2.getMemberFromString("Item in-rbi-rep:TwoMonths");
				allMembers.add(one);
				
				XBRLFactItem item = di.findFact(allMembers);
				System.out.println("The selected fact for member "+everyColumn+" is: "+item.toString());
				allMembers.remove( everyColumn );
			}

		} catch (LicenseVerificationException e) {
			e.printStackTrace();
		} catch (DTSDiscoveryException e) {
			e.printStackTrace();
		} catch (XBRLInitializationException e1) {
			e1.printStackTrace();
		} finally {
			// the end
		}
	}

}
