/*
 *  2017 by Ignacio Hernandez-Ros.
 *
 * This file is property of Reporting Standard S.L. and shall not be used outside the company unless you
 * have written permissions for different purposes.
 *
 * If you have received this file from Reporting Standard, you may have the following permissions:
 *
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2017 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 26 abr. 2017
 */
/**
 * 
 */
package com.ihr.xbrl.sample;

import java.io.File;
import java.net.URI;
import java.util.Properties;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLCoreProcessor;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.XMLDocumentURI;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.taxonomy.XBRLTaxonomy;
import com.ihr.xbrl.om.xdt.XDTProcessor;
import com.ihr.xbrl.util.TimeUtils;

/**
 * @author Ignacio
 *
 */
public class SampleTestStaticMode {

	/**
	 * @param args
	 * @throws Throwable 
	 */
	public static void main(String[] args) throws Throwable {
        System.out.println("Starting...");
        // the main purpose of this demo is to show how to obtain dimensions in the DTS.

        // we will:
        // 1. open a taxonomy that contains dimensions.
        // 2. validate the document against the XBRL 2.1 and Dimensions specifications.
        // 3. loop over the dimensions and show information about it.

        // Let's define the properties of the DTSContainer first. By default, the DTSContainer will perform only XBRL 2.1 validation.
        // we can easily configure the processor to perform XBRL 2.1 and Dimensions validation.

        Properties dtsProps = new Properties();
        dtsProps.setProperty(DTSContainer.PROCESSORS_SEQUENCE,
            XBRLCoreProcessor.PROCESSORNAME + "," + XDTProcessor.PROCESSORNAME);

        // Hint: adding more processors to the sequence is possible (Formulas) it is also possible creating custom
        // processors for validation of project specific things that are not part of the XBRL and cannot be part
        // of the Formulas, Dimensions or core XBRL 2.1 and attach them to the sequence.

        // Now lets create the URL of the taxonomy file with dimensions
        // change this if the file is in another directory.
        // The selected instance document is using COREP taxonomies 1.3.
        File fTaxonomy = new File("SamplesInCSharp/PerformanceTestsStaticMode/TaxonomyFiles/t-ca-2006-07-01.xsd");
        URI uTaxonomy = fTaxonomy.toURI(); // it is easier this way because URIs are more difficult to write when they are pointing to local files.

        // create the DTSContainer using indicated properties
        long start = System.currentTimeMillis();
        DTSContainer myDTS = DTSContainer.newEmptyContainer(dtsProps);
        long stop = System.currentTimeMillis();
        long duration = stop-start;

        System.out.println("Ready; "+ TimeUtils.toTime(duration));
//        System.out.println("Press any key to start loading the taxonomy...");
//        try {
//        	int count;
//        	while ((count = System.in.available()) > 0) {
//                System.in.skip(count);
//            }
//            System.in.read();
//        } catch(Exception e) {}  
        System.out.println("Loading... this may take a while...");

        // note this cast is not safe and will fail if the document pointed to by the URL is a linkbase or taxonomy. But this is a demo.
        start = System.currentTimeMillis();
        @SuppressWarnings("unused")
		XBRLTaxonomy myTaxonomy = (XBRLTaxonomy)myDTS.load(uTaxonomy);
        stop = System.currentTimeMillis();
        duration = stop - start;

        System.out.println("Loaded "+TimeUtils.toTime(duration)+"; Press any key to start validating the taxonomy...");
//        try {
//        	int count;
//        	while ((count = System.in.available()) > 0) {
//                System.in.skip(count);
//            }
//            System.in.read();
//        } catch(Exception e) {}  
        System.out.println("Validating...");

        // Now validate the instance document. This will loop all processor indicated in the PROCESSORS_SEQUENCE property
        // and will collect everything found (Errors, Inconsistencies, Assertions, Formula, ... into the XBRLValidationReport
        start = System.currentTimeMillis();
        XBRLValidationReport report = myDTS.validate();
        stop = System.currentTimeMillis();
        duration = stop - start;
        // in case of errors. we cannot continue. this call will thow the first error as an XBRLValidationExcecption and will do
        // nothing if the document is clean. Note we don't mind about inconsistencies at this stage. All of them are collected
        // inside the XBRLValidationReport
        report.throwExceptions();

        System.out.println("Validated " + TimeUtils.toTime(duration));
//        System.out.println("Press any key to start loading and validating the first instance document...");
//        try {
//        	int count;
//        	while ((count = System.in.available()) > 0) {
//                System.in.skip(count);
//            }
//            System.in.read();
//        } catch(Exception e) {}  

        // Now the DTS is ready to validate new instance documents under dynamic validation mode
        
        // Let's validate a small samle
        File fInstance1 = new File("SamplesInCSharp/PerformanceTestsStaticMode/TaxonomyFiles/CA_instance.xml");
        URI uInstance1 = fInstance1.toURI(); // it is easier this way because URIs are more difficult to write when they are pointing to local files.

        // load instance from disk
        start = System.currentTimeMillis();
        XBRLInstance instance1 = new XBRLInstance(new XMLDocumentURI(myDTS, uInstance1, null, false, null), false);
        XBRLInstance loadedInstance1 = myDTS.isInstanceOfThisDTS(instance1, DTSContainer.FAST); 
		if (loadedInstance1 == null)
		{
			System.out.println(fInstance1 + " Cannot be validated for this DTS.");
		}

        XBRLValidationReport rep = myDTS.validate(loadedInstance1);
        stop = System.currentTimeMillis();
        duration = stop - start;
        rep.throwExceptions();

        System.out.println("Validated " + TimeUtils.toTime(duration));
        loadedInstance1.clear(); // removes from the DTSContainer all information about this instance
//        System.out.println("Press any key to start loading and validating the second instance document...");
//        try {
//        	int count;
//        	while ((count = System.in.available()) > 0) {
//                System.in.skip(count);
//            }
//            System.in.read();
//        } catch(Exception e) {}  

        // Let's validate a large sample
        File fInstance2 = new File("SamplesInCSharp/PerformanceTestsStaticMode/TaxonomyFiles/CA_instanceLarge.xml");
        URI uInstance2 = fInstance2.toURI(); // it is easier this way because URIs are more difficult to write when they are pointing to local files.

        // load instance from disk
        start = System.currentTimeMillis();
        XBRLInstance instance2 = new XBRLInstance(new XMLDocumentURI(myDTS, uInstance2, null, false, null), false);
        XBRLInstance loadedInstance2 = myDTS.isInstanceOfThisDTS(instance2, DTSContainer.FAST);
        if (loadedInstance2 == null)
        {
            System.out.println(fInstance2 + " Cannot be validated for this DTS.");
        }

        XBRLValidationReport rep2 = myDTS.validate(loadedInstance2);
        loadedInstance2.clear();
        stop = System.currentTimeMillis();
        duration = stop - start;
        rep2.throwExceptions();

        System.out.println("Validated " + TimeUtils.toTime(duration));
        System.out.println("Finished; Press enter to exit...");
        try {
        	int count;
        	while ((count = System.in.available()) > 0) {
                System.in.skip(count);
            }
            System.in.read();
        } catch(Exception e) {}
        
	}

}
