/*
 *  2011 by Ignacio Hernandez-Ros.
 * This work may be reproduced and redistributed, in whole or in part, 
 * without alteration and without prior written permission, 
 * solely by educational institutions for nonprofit administrative 
 * or educational purposes provided all copies contain the 
 * following statement: 
 * " 2011 Ignacio Hernandez-Ros. This work is reproduced and distributed with the permission of Ignacio Hernandez-Ros. No other use is permitted without the express prior written permission of Ignacio Hernandez-Ros. For permission, contact ignacio@hernandez-ros.com"
 *
 * File creation date 14/09/2011
 */
/**
 * 
 */
package com.rs.xbrl.demo.validation;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Properties;

import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.DatatypeFactory;
import javax.xml.datatype.Duration;

import com.ihr.xbrl.om.DTSContainer;
import com.ihr.xbrl.om.XBRLDocument;
import com.ihr.xbrl.om.XBRLPlugInProcessor;
import com.ihr.xbrl.om.XBRLValidationReport;
import com.ihr.xbrl.om.XMLDocumentURI;
import com.ihr.xbrl.om.errors.XBRLValidationException;
import com.ihr.xbrl.om.instance.XBRLContext;
import com.ihr.xbrl.om.instance.XBRLInstance;
import com.ihr.xbrl.om.instance.XBRLPeriod;
import com.ihr.xbrl.util.GeneralIterable;
import com.rs.lic.LicenseVerificationException;

/**
 * @author Ignacio
 *
 */
public class DemoValidator implements XBRLPlugInProcessor {
	
	public static final String PROCESSORNAME = "DEMO Processor"; 
	
	public static final String ProcessorDescription = "This is a demo processor";

	// Used to hold the dts this processor will validate
	private DTSContainer dts;
	
	/**
	 * @param theDTS
	 */
	private DemoValidator(DTSContainer theDTS) {
		this.dts = theDTS;
		dts.registerProcessor(PROCESSORNAME,this);
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLPlugInProcessor#configure(java.util.Properties)
	 */
	@Override
	public void configure(Properties processorProperties) {
		
		// We will do nothing here.
		
		// The configure method is called on the DemoValidator processor instance when
		// the processor is created. The processorProperties contains a copy of all
		// properties passed in to the XBRL Processor. If this DemoValidator requires
		// external parameters, one possibility is to put external parameters into
		// properties and read the properties at this time.
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLPlugInProcessor#getDescription()
	 */
	@Override
	public String getDescription() {		
		// This is a string about the processor		
		return ProcessorDescription;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLPlugInProcessor#getName()
	 */
	@Override
	public String getName() {		
		// Returns the processor name. Used as a key to identify the processor.
		return PROCESSORNAME;
	}
	
	// Factory method to create instances of this processor by the processor factory
	public static DemoValidator create(DTSContainer dts) throws LicenseVerificationException {
		if (dts == null)
			return null;
		DemoValidator proc = dts.getProcessor(PROCESSORNAME);
		if (proc != null) {
			return proc;
		}
		
		proc = new DemoValidator(dts);
		return proc;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLPlugInProcessor#validate()
	 */
	@Override
	public XBRLValidationReport validate() {
		// This processor will only validate XBRL Reports, so, obtain the report from the dts
		// and call validate(XBRLDocument)
		XMLDocumentURI doc = dts.cacheGet(dts.getRootURI());
		if (doc instanceof XBRLInstance)
			return validate((XBRLInstance)doc);
		return null;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLPlugInProcessor#validate(com.ihr.xbrl.om.XBRLDocument)
	 */
	@Override
	public XBRLValidationReport validate(XBRLDocument report) {
		XBRLValidationReport valReport = new XBRLValidationReport(dts);
		if (report instanceof XBRLInstance) {
			
			// here is the code that performs validation
			
			XBRLInstance instance = (XBRLInstance)report;
			
			// loop over all XBRL Contexts in the instance
			for (XBRLContext ctx : GeneralIterable.asIterator( instance.getInstanceRootNode().getContexts() )) {
				XBRLPeriod p = ctx.getPeriod();
				// if ctx is duration, check the duration period is no longer than a year
				if (p.isDuration()) {
					try {
						Duration d = calculateDuration(p.getGregorianCalendarDate(XBRLPeriod.DATE_START), p.getGregorianCalendarDate(XBRLPeriod.DATE_END));
						if (d.isLongerThan(DatatypeFactory.newInstance().newDuration("P1Y"))) {
							// add the error to the XBRL Validation report
							
							valReport.addException(new XBRLValidationException(333,"At context "+ctx.getId()+" the period is longer than 1 year."));
						}
					} catch (DatatypeConfigurationException e) {
						// this is an internal application error
						throw new RuntimeException("Internal error inside validate(XBRLReport)",e);
					}
				}
				
			}
		}
		return valReport;
	}
	
	/**
	 * @param start
	 * @param end
	 * @return
	 * @throws DatatypeConfigurationException 
	 */
	private static Duration calculateDuration(GregorianCalendar start,	GregorianCalendar end) throws DatatypeConfigurationException {
		if (end == null && start == null)
			return null;
		int years = end.get(Calendar.YEAR)-start.get(Calendar.YEAR);
		int months = end.get(Calendar.MONTH)-start.get(Calendar.MONTH);
		if (months < 0) {
			months = 12 + months;
			--years;
		}
		int days = end.get(Calendar.DAY_OF_MONTH)-start.get(Calendar.DAY_OF_MONTH);
		if (days < 0) {
			days = 0; // TODO: this is like a bug... but is not so impotant for the time being
		}
		int hours = end.get(Calendar.HOUR)-start.get(Calendar.HOUR);
		if (hours < 0) {
			hours = 0;
		}
		int minutes = end.get(Calendar.MINUTE)-start.get(Calendar.MINUTE);
		if (minutes < 0) {
			minutes = 0;
		}
		int seconds = end.get(Calendar.SECOND)-start.get(Calendar.SECOND);
		if (seconds < 0) {
			seconds = 0;
		}
		
		Duration d = DatatypeFactory.newInstance().newDuration(true,
				years,
				months,
				days,
				hours,
				minutes,
				seconds
				);
		
		return d;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLPlugInProcessor#isHidden()
	 */
	@Override
	public boolean isHidden() {
		return false;
	}

	/* (non-Javadoc)
	 * @see com.ihr.xbrl.om.XBRLPlugInProcessor#clear()
	 */
	@Override
	public void clear() {
		dts = null;
	}
}

//
// This software is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Ignacio Hernandez-Ros.
//
// Contributor(s): none.
